import asyncio

from asynctnt import Response
from tests import BaseTarantoolTestCase
from tests.util import get_complex_param


class EvalTestCase(BaseTarantoolTestCase):
    async def test__eval_basic(self):
        res = await self.conn.eval('return "hola"')

        self.assertIsInstance(res, Response, "Got eval response")
        self.assertEqual(res.code, 0, "success")
        self.assertGreater(res.sync, 0, "sync > 0")
        self.assertResponseEqual(res, ["hola"], "Body ok")

    async def test__eval_basic_pack(self):
        res = await self.conn.eval('return {"hola"}')

        self.assertIsInstance(res, Response, "Got eval response")
        self.assertEqual(res.code, 0, "success")
        self.assertGreater(res.sync, 0, "sync > 0")
        self.assertResponseEqual(res, [["hola"]], "Body ok")

    async def test__eval_with_param(self):
        args = [1, 2, 3, "hello"]
        res = await self.conn.eval("return ...", args)

        self.assertResponseEqual(res, args, "Body ok")

    async def test__eval_with_param_pack(self):
        args = [1, 2, 3, "hello"]
        res = await self.conn.eval("return {...}", args)

        self.assertResponseEqual(res, [args], "Body ok")

    async def test__eval_func_name_invalid_type(self):
        with self.assertRaises(TypeError):
            await self.conn.eval(12)

        with self.assertRaises(TypeError):
            await self.conn.eval([1, 2])

        with self.assertRaises(TypeError):
            await self.conn.eval({"a": 1})

        with self.assertRaises(TypeError):
            await self.conn.eval(b"qwer")

    async def test__eval_params_invalid_type(self):
        with self.assertRaises(TypeError):
            await self.conn.eval("return {...}", 220349)

        with self.assertRaises(TypeError):
            await self.conn.eval("return {...}", "hey")

        with self.assertRaises(TypeError):
            await self.conn.eval("return {...}", {1: 1, 2: 2})

    async def test__eval_args_tuple(self):
        try:
            await self.conn.eval("return {...}", (1, 2))
        except Exception as e:
            self.fail(e)

    async def test__eval_complex_param(self):
        p, cmp = get_complex_param(
            encoding=self.conn.encoding, replace_bin=self.conn.version < (3, 0)
        )
        res = await self.conn.eval("return {...}", [p])
        self.assertDictEqual(res[0][0], cmp, "Body ok")

    async def test__eval_timeout_in_time(self):
        try:
            cmd = """
            local args = {...}
            local fiber = require("fiber")
            fiber.sleep(args[1])
            """
            await self.conn.eval(cmd, [0.1], timeout=1)
        except Exception as e:
            self.fail(e)

    async def test__eval_timeout_late(self):
        cmd = """
        local args = {...}
        local fiber = require("fiber")
        fiber.sleep(args[1])
        """
        with self.assertRaises(asyncio.TimeoutError):
            await self.conn.eval(cmd, [0.3], timeout=0.1)



import asyncio
import uuid

import asynctnt
from asynctnt.connection import ConnectionState
from asynctnt.exceptions import (
    ErrorCode,
    TarantoolDatabaseError,
    TarantoolNotConnectedError,
)
from asynctnt.instance import TarantoolSyncInstance
from tests import BaseTarantoolTestCase
from tests._testbase import check_version


class ConnectTestCase(BaseTarantoolTestCase):
    DO_CONNECT = False

    async def test__connect(self):
        conn = asynctnt.Connection(
            host=self.tnt.host, port=self.tnt.port, reconnect_timeout=0
        )
        self.assertEqual(conn.host, self.tnt.host)
        self.assertEqual(conn.port, self.tnt.port)
        self.assertIsNone(conn.username)
        self.assertIsNone(conn.password)
        self.assertEqual(conn.reconnect_timeout, 0)
        self.assertEqual(conn.connect_timeout, 3)
        self.assertIsNone(conn.initial_read_buffer_size)
        self.assertIsNone(conn.schema_id)
        self.assertIsNone(conn.version)
        self.assertEqual(
            repr(conn),
            "<asynctnt.Connection host={} port={} state={}>".format(
                conn.host, conn.port, repr(conn.state)
            ),
        )
        self.assertEqual(conn.state, ConnectionState.DISCONNECTED)

        c = await conn.connect()
        self.assertEqual(c, conn)
        self.assertIsNotNone(conn._transport)
        self.assertIsNotNone(conn._protocol)
        self.assertTrue(conn.is_connected)
        self.assertTrue(conn.is_fully_connected)
        self.assertEqual(conn.state, ConnectionState.CONNECTED)
        self.assertIsNotNone(conn._protocol.schema)
        self.assertIsNotNone(conn.version)

        self.assertIsNotNone(conn.schema.id)
        self.assertIsNotNone(conn.schema.spaces)

        await conn.call("box.info")
        await conn.disconnect()

    async def test__connect_direct(self):
        conn = await asynctnt.connect(
            host=self.tnt.host, port=self.tnt.port, reconnect_timeout=0
        )
        self.assertEqual(conn.host, self.tnt.host)
        self.assertEqual(conn.port, self.tnt.port)
        self.assertIsNone(conn.username)
        self.assertIsNone(conn.password)
        self.assertEqual(conn.reconnect_timeout, 0)
        self.assertEqual(conn.connect_timeout, 3)
        self.assertIsNone(conn.initial_read_buffer_size)

        self.assertIsNotNone(conn._transport)
        self.assertIsNotNone(conn._protocol)
        self.assertTrue(conn.is_connected)
        self.assertTrue(conn.is_fully_connected)
        self.assertEqual(conn.state, ConnectionState.CONNECTED)
        self.assertIsNotNone(conn._protocol.schema)
        self.assertIsNotNone(conn.version)

        await conn.call("box.info")
        await conn.disconnect()

    async def test__connect_unix(self):
        if self.in_docker:
            self.skipTest("Skipping as running inside the docker")
            return

        tnt = TarantoolSyncInstance(
            host="unix/",
            port="/tmp/" + uuid.uuid4().hex + ".sock",
            console_host="127.0.0.1",
            applua=self.read_applua(),
            cleanup=self.TNT_CLEANUP,
        )
        tnt.start()
        try:
            conn = await asynctnt.connect(
                host=tnt.host, port=tnt.port, reconnect_timeout=0
            )
            self.assertEqual(conn.host, tnt.host)
            self.assertEqual(conn.port, tnt.port)
            self.assertIsNone(conn.username)
            self.assertIsNone(conn.password)
            self.assertEqual(conn.reconnect_timeout, 0)
            self.assertEqual(conn.connect_timeout, 3)
            self.assertIsNone(conn.initial_read_buffer_size)

            self.assertIsNotNone(conn._transport)
            self.assertIsNotNone(conn._protocol)
            self.assertTrue(conn.is_connected)
            self.assertTrue(conn.is_fully_connected)
            self.assertEqual(conn.state, ConnectionState.CONNECTED)
            self.assertIsNotNone(conn._protocol.schema)
            self.assertIsNotNone(conn.version)

            await conn.call("box.info")
            await conn.disconnect()
        finally:
            tnt.stop()

    async def test__connect_contextmanager(self):
        conn = asynctnt.Connection(
            host=self.tnt.host, port=self.tnt.port, reconnect_timeout=0
        )
        self.assertEqual(conn.state, ConnectionState.DISCONNECTED)

        async with conn:
            self.assertIsNotNone(conn._transport)
            self.assertIsNotNone(conn._protocol)
            self.assertTrue(conn.is_connected)
            self.assertTrue(conn.is_fully_connected)
            self.assertEqual(conn.state, ConnectionState.CONNECTED)
            self.assertIsNotNone(conn._protocol.schema)
            self.assertIsNotNone(conn.version)

            await conn.call("box.info")

        self.assertEqual(conn.state, ConnectionState.DISCONNECTED)

    async def test__connect_contextmanager_connect_inside(self):
        conn = asynctnt.Connection(
            host=self.tnt.host, port=self.tnt.port, reconnect_timeout=0
        )
        self.assertEqual(conn.state, ConnectionState.DISCONNECTED)

        async with conn:
            await conn.connect()
            self.assertEqual(conn.state, ConnectionState.CONNECTED)
            await conn.call("box.info")

        self.assertEqual(conn.state, ConnectionState.DISCONNECTED)

    async def test__connect_contextmanager_disconnect_inside(self):
        conn = asynctnt.Connection(
            host=self.tnt.host, port=self.tnt.port, reconnect_timeout=0
        )
        self.assertEqual(conn.state, ConnectionState.DISCONNECTED)

        async with conn:
            await conn.disconnect()
            self.assertEqual(conn.state, ConnectionState.DISCONNECTED)

            with self.assertRaises(TarantoolNotConnectedError):
                await conn.call("box.info")

        self.assertEqual(conn.state, ConnectionState.DISCONNECTED)

    async def test__connect_no_schema(self):
        conn = asynctnt.Connection(
            host=self.tnt.host,
            port=self.tnt.port,
            reconnect_timeout=0,
            fetch_schema=False,
            auto_refetch_schema=False,
        )
        async with conn:
            self.assertIsNotNone(conn._transport)
            self.assertIsNotNone(conn._protocol)
            self.assertTrue(conn.is_connected)
            self.assertTrue(conn.is_fully_connected)
            self.assertEqual(conn.state, ConnectionState.CONNECTED)
            self.assertIsNotNone(conn._protocol.schema)
            await conn.call("box.info")

    async def test__connect_auth(self):
        conn = asynctnt.Connection(
            host=self.tnt.host,
            port=self.tnt.port,
            username="t1",
            password="t1",
            reconnect_timeout=0,
        )
        async with conn:
            self.assertIsNotNone(conn._transport)
            self.assertIsNotNone(conn._protocol)
            self.assertTrue(conn.is_connected)
            self.assertTrue(conn.is_fully_connected)
            self.assertEqual(conn.state, ConnectionState.CONNECTED)
            self.assertIsNotNone(conn._protocol.schema)
            await conn.call("box.info")

    async def test__connect_auth_no_schema(self):
        conn = asynctnt.Connection(
            host=self.tnt.host,
            port=self.tnt.port,
            username="t1",
            password="t1",
            fetch_schema=False,
            auto_refetch_schema=False,
            reconnect_timeout=0,
        )
        self.assertEqual(conn.username, "t1")
        self.assertEqual(conn.password, "t1")
        async with conn:
            self.assertIsNotNone(conn._transport)
            self.assertIsNotNone(conn._protocol)
            self.assertTrue(conn.is_connected)
            self.assertTrue(conn.is_fully_connected)
            self.assertEqual(conn.state, ConnectionState.CONNECTED)
            self.assertIsNotNone(conn._protocol.schema)
            await conn.call("box.info")

    async def test__disconnect(self):
        conn = asynctnt.Connection(
            host=self.tnt.host, port=self.tnt.port, reconnect_timeout=0
        )
        await conn.connect()
        await conn.disconnect()
        self.assertFalse(conn.is_connected)
        self.assertFalse(conn.is_fully_connected)
        self.assertEqual(conn.state, ConnectionState.DISCONNECTED)

        with self.assertRaises(TarantoolNotConnectedError):
            await conn.call("box.info")

    async def test__disconnect_in_request(self):
        conn = asynctnt.Connection(
            host=self.tnt.host, port=self.tnt.port, reconnect_timeout=0
        )
        await conn.connect()

        coro = self.ensure_future(conn.eval('require "fiber".sleep(2)'))
        await self.sleep(0.5)
        await conn.disconnect()

        with self.assertRaises(TarantoolNotConnectedError):
            await coro

    async def test__disconnect_auth(self):
        conn = asynctnt.Connection(
            host=self.tnt.host,
            port=self.tnt.port,
            username="t1",
            password="t1",
            reconnect_timeout=0,
        )
        await conn.connect()
        await conn.disconnect()
        self.assertFalse(conn.is_connected)
        self.assertFalse(conn.is_fully_connected)
        self.assertEqual(conn.state, ConnectionState.DISCONNECTED)

        with self.assertRaises(TarantoolNotConnectedError):
            await conn.call("box.info")

    async def test__disconnect_while_reconnecting(self):
        conn = asynctnt.Connection(
            host=self.tnt.host,
            port=self.tnt.port,
            username="t1",
            password="t1",
            reconnect_timeout=0.1,
        )
        self.assertEqual(conn.reconnect_timeout, 0.1)
        try:
            await conn.connect()
            self.tnt.stop()
            await self.sleep(0.5)

            await conn.disconnect()

            self.assertFalse(conn.is_connected)
            self.assertFalse(conn.is_fully_connected)
            self.assertEqual(conn.state, ConnectionState.DISCONNECTED)

            with self.assertRaises(TarantoolNotConnectedError):
                await conn.call("box.info")
        finally:
            self.tnt.start()

    async def test__close_while_reconnecting(self):
        conn = asynctnt.Connection(
            host=self.tnt.host,
            port=self.tnt.port,
            username="t1",
            password="t1",
            reconnect_timeout=0.1,
        )
        self.assertEqual(conn.reconnect_timeout, 0.1)
        try:
            await conn.connect()
            self.tnt.stop()
            await self.sleep(0.5)

            conn.close()

            self.assertFalse(conn.is_connected)
            self.assertFalse(conn.is_fully_connected)
            self.assertEqual(conn.state, ConnectionState.DISCONNECTED)

            with self.assertRaises(TarantoolNotConnectedError):
                await conn.call("box.info")
        finally:
            self.tnt.start()

    async def test__connect_multiple(self):
        conn = asynctnt.Connection(
            host=self.tnt.host,
            port=self.tnt.port,
            fetch_schema=False,
            reconnect_timeout=0,
        )
        for _ in range(10):
            await conn.connect()
            await conn.disconnect()
        self.assertFalse(conn.is_connected)
        self.assertFalse(conn.is_fully_connected)
        self.assertEqual(conn.state, ConnectionState.DISCONNECTED)

        with self.assertRaises(TarantoolNotConnectedError):
            await conn.call("box.info")

    async def test__connect_cancel(self):
        conn = asynctnt.Connection(
            host=self.tnt.host,
            port=self.tnt.port,
            fetch_schema=True,
            reconnect_timeout=0,
        )
        try:
            f = asyncio.ensure_future(conn.connect())
            await self.sleep(0.0001)
            f.cancel()
            with self.assertRaises(asyncio.CancelledError):
                await f
        finally:
            await conn.disconnect()

    async def test__connect_error_no_reconnect(self):
        conn = asynctnt.Connection(
            host="127.0.0.1", port=1, fetch_schema=True, reconnect_timeout=0
        )
        with self.assertRaises(ConnectionRefusedError):
            await conn.connect()

    async def test__connect_wait_tnt_started(self):
        self.tnt.stop()
        conn = asynctnt.Connection(
            host=self.tnt.host,
            port=self.tnt.port,
            username="t1",
            password="t1",
            fetch_schema=True,
            reconnect_timeout=0.000000001,
        )
        try:
            coro = self.ensure_future(conn.connect())
            await self.sleep(0.3)
            self.tnt.start()
            await self.sleep(1)
            while True:
                try:
                    await coro
                    break
                except TarantoolDatabaseError as e:
                    if e.code == ErrorCode.ER_NO_SUCH_USER:
                        # Try again
                        coro = self.ensure_future(conn.connect())
                        continue
                    raise

            self.assertEqual(conn.state, ConnectionState.CONNECTED)

            await conn.call("box.info")
        finally:
            await conn.disconnect()

    async def test__connect_waiting_for_spaces(self):
        if self.in_docker:
            self.skipTest("not running in docker")
            return

        with self.make_instance() as tnt:
            tnt.replication_source = ["x:1"]
            tnt.start(wait=False)

            conn = asynctnt.Connection(
                host=tnt.host,
                port=tnt.port,
                fetch_schema=True,
                reconnect_timeout=0.1,
                connect_timeout=10,
            )
            self.assertEqual(conn.connect_timeout, 10)
            try:
                states = {}

                async def state_checker():
                    while True:
                        states[conn.state] = True
                        await self.sleep(0.001)

                checker = self.ensure_future(state_checker())

                try:
                    await asyncio.wait_for(conn.connect(), 1)
                except asyncio.TimeoutError:
                    self.assertTrue(True, "connect cancelled")

                checker.cancel()

                self.assertTrue(
                    states.get(ConnectionState.CONNECTING, False), "was in connecting"
                )

                with self.assertRaises(TarantoolNotConnectedError):
                    await conn.call("box.info")
            finally:
                await conn.disconnect()

    async def test__connect_waiting_for_spaces_no_reconnect(self):
        if self.in_docker:
            self.skipTest("not running in docker")
            return

        with self.make_instance() as tnt:
            tnt.replication_source = ["x:1"]
            tnt.start(wait=False)
            await self.sleep(1)

            if not check_version(self, tnt.version(), min=(1, 7)):
                return

            conn = asynctnt.Connection(
                host=tnt.host,
                port=tnt.port,
                fetch_schema=True,
                reconnect_timeout=0,
                connect_timeout=10,
            )
            try:
                with self.assertRaises(TarantoolDatabaseError) as e:
                    await conn.connect()

                self.assertEqual(e.exception.code, ErrorCode.ER_NO_SUCH_SPACE)
            finally:
                await conn.disconnect()

    async def test__connect_waiting_for_spaces_no_reconnect_1_6(self):
        with self.make_instance() as tnt:
            tnt.replication_source = ["x:1"]
            tnt.start(wait=False)
            await self.sleep(1)

            if not check_version(self, tnt.version(), max=(1, 7)):
                return

            conn = asynctnt.Connection(
                host=tnt.host,
                port=tnt.port,
                fetch_schema=True,
                reconnect_timeout=0,
                connect_timeout=10,
            )
            try:
                with self.assertRaises(ConnectionRefusedError):
                    await conn.connect()
            finally:
                await conn.disconnect()

    async def test__connect_err_loading(self):
        if self.in_docker:
            self.skipTest("not running in docker")
            return

        with self.make_instance() as tnt:
            tnt.replication_source = ["x:1"]
            tnt.start(wait=False)
            await self.sleep(1)

            if not check_version(self, tnt.version(), min=(1, 7)):
                return

            conn = asynctnt.Connection(
                host=tnt.host,
                port=tnt.port,
                username="t1",
                password="t1",
                fetch_schema=True,
                reconnect_timeout=0,
                connect_timeout=10,
            )
            try:
                with self.assertRaises(TarantoolDatabaseError) as e:
                    await conn.connect()

                self.assertEqual(e.exception.code, ErrorCode.ER_LOADING)
            finally:
                await conn.disconnect()

    async def test__connect_err_loading_1_6(self):
        with self.make_instance() as tnt:
            tnt.replication_source = ["x:1"]
            tnt.start(wait=False)

            await self.sleep(1)
            if not check_version(self, tnt.version(), max=(1, 7)):
                return

            conn = asynctnt.Connection(
                host=tnt.host,
                port=tnt.port,
                username="t1",
                password="t1",
                fetch_schema=True,
                reconnect_timeout=0,
                connect_timeout=10,
            )
            try:
                with self.assertRaises(ConnectionRefusedError):
                    await conn.connect()
            finally:
                await conn.disconnect()

    async def test__connect_tnt_restarted(self):
        conn = asynctnt.Connection(
            host=self.tnt.host,
            port=self.tnt.port,
            username="t1",
            password="t1",
            fetch_schema=True,
            reconnect_timeout=0.000001,
        )
        await conn.connect()

        try:
            self.tnt.stop()
            self.tnt.start()
            await self.sleep(0.5)
            await conn.ping()
        except Exception as e:
            self.fail("Should not throw any exceptions, but got: {}".format(e))
        finally:
            await conn.disconnect()

    async def test__connect_force_disconnect(self):
        conn = asynctnt.Connection(
            host=self.tnt.host, port=44444, reconnect_timeout=0.3
        )
        self.ensure_future(conn.connect())
        await self.sleep(1)
        await conn.disconnect()
        self.assertEqual(conn.state, ConnectionState.DISCONNECTED)

    async def test__close(self):
        conn = asynctnt.Connection(
            host=self.tnt.host, port=self.tnt.port, reconnect_timeout=0
        )
        await conn.connect()
        await self.sleep(0.1)
        conn.close()
        await self.sleep(0.1)
        self.assertEqual(conn.state, ConnectionState.DISCONNECTED)

    async def test_disconnect_from_idle(self):
        conn = asynctnt.Connection(
            host=self.tnt.host, port=self.tnt.port, reconnect_timeout=0
        )
        await conn.disconnect()
        self.assertEqual(conn.state, ConnectionState.DISCONNECTED)

    async def test_reconnect_from_idle(self):
        conn = asynctnt.Connection(
            host=self.tnt.host, port=self.tnt.port, reconnect_timeout=0
        )
        await conn.reconnect()
        try:
            self.assertEqual(conn.state, ConnectionState.CONNECTED)
            await conn.call("box.info")
        finally:
            await conn.disconnect()

    async def test_reconnect_after_connect(self):
        conn = asynctnt.Connection(
            host=self.tnt.host, port=self.tnt.port, reconnect_timeout=0
        )
        try:
            await conn.connect()
            await conn.reconnect()

            self.assertEqual(conn.state, ConnectionState.CONNECTED)
            await conn.call("box.info")
        finally:
            await conn.disconnect()

    async def test_manual_reconnect(self):
        conn = asynctnt.Connection(
            host=self.tnt.host, port=self.tnt.port, reconnect_timeout=0
        )
        try:
            await conn.connect()
            await conn.disconnect()
            await conn.connect()

            self.assertEqual(conn.state, ConnectionState.CONNECTED)
            await conn.call("box.info")
        finally:
            await conn.disconnect()

    async def test__connect_connection_lost(self):
        conn = asynctnt.Connection(
            host=self.tnt.host, port=self.tnt.port, reconnect_timeout=1 / 3
        )
        try:
            await conn.connect()
            self.tnt.stop()
            await self.sleep(0.5)
            self.tnt.start()
            await self.sleep(0.5)

            self.assertEqual(conn.state, ConnectionState.CONNECTED)
            self.assertTrue(conn.is_connected)
            await conn.call("box.info")
        finally:
            await conn.disconnect()

    async def test__connect_from_multiple_coroutines(self):
        conn = asynctnt.Connection(
            host=self.tnt.host, port=self.tnt.port, reconnect_timeout=1 / 3
        )
        try:
            coros = []
            for _ in range(10):
                coros.append(asyncio.ensure_future(conn.connect()))

            await asyncio.gather(*coros)
            self.assertEqual(conn.state, ConnectionState.CONNECTED)
            self.assertTrue(conn.is_connected)
            await conn.call("box.info")
        finally:
            await conn.disconnect()

    async def test__disconnect_from_multiple_coroutines(self):
        conn = asynctnt.Connection(
            host=self.tnt.host, port=self.tnt.port, reconnect_timeout=1 / 3
        )
        try:
            await conn.connect()
            coros = []
            for _ in range(10):
                coros.append(asyncio.ensure_future(conn.disconnect()))

            await asyncio.gather(*coros)
            self.assertEqual(conn.state, ConnectionState.DISCONNECTED)
            self.assertFalse(conn.is_connected)

            with self.assertRaises(TarantoolNotConnectedError):
                await conn.call("box.info")
        finally:
            await conn.disconnect()

    async def test__connect_while_reconnecting(self):
        conn = asynctnt.Connection(
            host=self.tnt.host, port=self.tnt.port, reconnect_timeout=1
        )

        try:
            coros = []
            for _ in range(10):
                coros.append(asyncio.ensure_future(conn.connect()))

            self.tnt.stop()
            await self.sleep(0.5)

            connect_coros = asyncio.ensure_future(asyncio.gather(*coros))

            self.tnt.start()
            await self.sleep(1)
            await connect_coros

            self.assertEqual(conn.state, ConnectionState.CONNECTED)
            self.assertTrue(conn.is_connected)

            await conn.call("box.info")
        finally:
            await conn.disconnect()

    async def test__connect_on_tnt_crash_no_reconnect(self):
        conn = asynctnt.Connection(
            host=self.tnt.host,
            port=self.tnt.port,
            connect_timeout=1,
            reconnect_timeout=0,
        )
        try:
            await conn.connect()
            await asyncio.sleep(0.5)
            try:
                await conn.eval("require('ffi').cast('char *', 0)[0] = 48")
            except TarantoolNotConnectedError:
                self.assertTrue(True, "not connected error triggered")
            self.tnt.stop()
            self.tnt.start()
            await asyncio.sleep(1)
            await conn.connect()  # this connect should reconnect easily

            await conn.call("box.info")
        finally:
            await conn.disconnect()

    async def test__connect_on_tnt_crash_with_reconnect(self):
        conn = asynctnt.Connection(
            host=self.tnt.host,
            port=self.tnt.port,
            connect_timeout=1,
            reconnect_timeout=1 / 3,
        )
        try:
            await conn.connect()
            await asyncio.sleep(0.5)
            try:
                await conn.eval("require('ffi').cast('char *', 0)[0] = 48")
            except TarantoolNotConnectedError:
                self.assertTrue(True, "not connected error triggered")
            self.tnt.stop()
            self.tnt.start()
            await asyncio.sleep(1)
            await conn.connect()  # this connect should reconnect easily

            await conn.call("box.info")
        finally:
            await conn.disconnect()

    async def test__connect_invalid_user_no_reconnect(self):
        async with asynctnt.Connection(host=self.tnt.host, port=self.tnt.port) as conn:
            version = conn.version

        conn = asynctnt.Connection(
            host=self.tnt.host,
            port=self.tnt.port,
            username="fancy",
            password="man",
            connect_timeout=1,
            reconnect_timeout=0,
        )
        with self.assertRaises(TarantoolDatabaseError) as e:
            await conn.connect()

        err_code = ErrorCode.ER_PASSWORD_MISMATCH
        if version < (2, 11):
            err_code = ErrorCode.ER_NO_SUCH_USER
        self.assertEqual(e.exception.code, err_code)

    async def test__connect_invalid_user_with_reconnect(self):
        conn = asynctnt.Connection(
            host=self.tnt.host,
            port=self.tnt.port,
            fetch_schema=True,
            reconnect_timeout=0.1,
            connect_timeout=10,
        )
        await conn.connect()  # first connect successfully

        # then change credentials
        conn._username = "fancy"
        conn._password = "man"

        self.tnt.stop()
        self.tnt.start()
        await self.sleep(0.1)
        try:
            states = {}

            async def state_checker():
                while True:
                    states[conn.state] = True
                    await self.sleep(0.001)

            checker = self.ensure_future(state_checker())

            try:
                await asyncio.wait_for(conn.connect(), 1)
            except asyncio.TimeoutError:
                self.assertTrue(True, "connect cancelled")

            checker.cancel()

            self.assertTrue(
                states.get(ConnectionState.CONNECTING, False), "was in connecting"
            )
            self.assertTrue(
                states.get(ConnectionState.RECONNECTING, False), "was in connecting"
            )

            with self.assertRaises(TarantoolNotConnectedError):
                await conn.call("box.info")
        finally:
            await conn.disconnect()



import asyncio

from asynctnt import Response
from asynctnt.exceptions import ErrorCode, TarantoolDatabaseError
from tests import BaseTarantoolTestCase
from tests.util import get_complex_param


class CallTestCase(BaseTarantoolTestCase):
    def has_new_call(self):
        return self.conn.version >= (1, 7)

    async def test__call_basic(self):
        res = await self.conn.call("func_hello")

        self.assertIsInstance(res, Response, "Got call response")
        self.assertEqual(res.code, 0, "success")
        self.assertGreater(res.sync, 0, "sync > 0")
        self.assertResponseEqual(res, [["hello"]], "Body ok")

    async def test__call_basic_bare(self):
        res = await self.conn.call("func_hello_bare")
        cmp = ["hello"]

        self.assertIsInstance(res, Response, "Got call response")
        self.assertEqual(res.code, 0, "success")
        self.assertGreater(res.sync, 0, "sync > 0")
        if not self.has_new_call():
            cmp = [cmp]
        self.assertResponseEqual(res, cmp, "Body ok")

    async def test__call_unknown_function(self):
        with self.assertRaises(TarantoolDatabaseError) as ctx:
            await self.conn.call("blablabla")
        self.assertEqual(ctx.exception.code, ErrorCode.ER_NO_SUCH_PROC)

    async def test__call_with_param(self):
        res = await self.conn.call("func_param", ["myparam"])

        self.assertIsInstance(res, Response, "Got call response")
        self.assertResponseEqual(res, [["myparam"]], "Body ok")

    async def test__call_with_param_bare(self):
        res = await self.conn.call("func_param_bare", ["myparam"])
        cmp = ["myparam"]
        if not self.has_new_call():
            cmp = [cmp]

        self.assertIsInstance(res, Response, "Got call response")
        self.assertResponseEqual(res, cmp, "Body ok")

    async def test__call_func_name_invalid_type(self):
        with self.assertRaises(TypeError):
            await self.conn.call(12)

        with self.assertRaises(TypeError):
            await self.conn.call([1, 2])

        with self.assertRaises(TypeError):
            await self.conn.call({"a": 1})

        with self.assertRaises(TypeError):
            await self.conn.call(b"qwer")

    async def test__call_params_invalid_type(self):
        with self.assertRaises(TypeError):
            await self.conn.call("func_param", 220349)

        with self.assertRaises(TypeError):
            await self.conn.call("func_param", "hey")

        with self.assertRaises(TypeError):
            await self.conn.call("func_param", {1: 1, 2: 2})

    async def test__call_args_tuple(self):
        try:
            await self.conn.call("func_param", (1, 2))
        except Exception as e:
            self.fail(e)

    async def test__call_complex_param(self):
        p, cmp = get_complex_param(
            encoding=self.conn.encoding, replace_bin=self.conn.version < (3, 0)
        )
        res = await self.conn.call("func_param", [p])
        self.assertDictEqual(res[0][0], cmp, "Body ok")

    async def test__call_complex_param_bare(self):
        p, cmp = get_complex_param(
            encoding=self.conn.encoding, replace_bin=self.conn.version < (3, 0)
        )
        cmp = [cmp]
        res = await self.conn.call("func_param_bare", [p])
        if not self.has_new_call():
            cmp = [cmp]
        self.assertResponseEqual(res, cmp, "Body ok")

    async def test__call_timeout_in_time(self):
        try:
            await self.conn.call("func_long", [0.1], timeout=1)
        except Exception as e:
            self.fail(e)

    async def test__call_timeout_late(self):
        with self.assertRaises(asyncio.TimeoutError):
            await self.conn.call("func_long", [0.3], timeout=0.1)

    async def test__call_raise(self):
        with self.assertRaises(TarantoolDatabaseError) as e:
            await self.conn.call("raise")

        self.assertEqual(e.exception.code, 0, "code by box.error{} is 0")
        self.assertEqual(e.exception.message, "my reason", "Reason ok")


class Call16TestCase(BaseTarantoolTestCase):
    async def test__call16_basic(self):
        res = await self.conn.call16("func_hello")

        self.assertIsInstance(res, Response, "Got call response")
        self.assertEqual(res.code, 0, "success")
        self.assertGreater(res.sync, 0, "sync > 0")
        self.assertResponseEqual(res, [["hello"]], "Body ok")

    async def test__call16_basic_bare(self):
        # Tarantool automatically wraps return result into tuple

        res = await self.conn.call16("func_hello")

        self.assertIsInstance(res, Response, "Got call response")
        self.assertEqual(res.code, 0, "success")
        self.assertGreater(res.sync, 0, "sync > 0")
        self.assertResponseEqual(res, [["hello"]], "Body ok")

    async def test__call16_unknown_function(self):
        with self.assertRaises(TarantoolDatabaseError) as ctx:
            await self.conn.call16("blablabla")
        self.assertEqual(ctx.exception.code, ErrorCode.ER_NO_SUCH_PROC)

    async def test__call16_with_param(self):
        res = await self.conn.call16("func_param", ["myparam"])

        self.assertIsInstance(res, Response, "Got call response")
        self.assertResponseEqual(res, [["myparam"]], "Body ok")

    async def test__call16_with_param_bare(self):
        # Tarantool automatically wraps return result into tuple

        res = await self.conn.call16("func_param_bare", ["myparam"])

        self.assertIsInstance(res, Response, "Got call response")
        self.assertResponseEqual(res, [["myparam"]], "Body ok")

    async def test__call16_func_name_invalid_type(self):
        with self.assertRaises(TypeError):
            await self.conn.call16(12)

        with self.assertRaises(TypeError):
            await self.conn.call16([1, 2])

        with self.assertRaises(TypeError):
            await self.conn.call16({"a": 1})

        with self.assertRaises(TypeError):
            await self.conn.call16(b"qwer")

    async def test__call16_params_invalid_type(self):
        with self.assertRaises(TypeError):
            await self.conn.call16("func_param", 220349)

        with self.assertRaises(TypeError):
            await self.conn.call16("func_param", "hey")

        with self.assertRaises(TypeError):
            await self.conn.call16("func_param", {1: 1, 2: 2})

    async def test__call16_args_tuple(self):
        try:
            await self.conn.call16("func_param", (1, 2))
        except Exception as e:
            self.fail(e)

    async def test__call16_complex_param(self):
        p, cmp = get_complex_param(
            encoding=self.conn.encoding, replace_bin=self.conn.version < (3, 0)
        )
        res = await self.conn.call("func_param", [p])
        self.assertDictEqual(res[0][0], cmp, "Body ok")

    async def test__call16_complex_param_bare(self):
        p, cmp = get_complex_param(
            encoding=self.conn.encoding, replace_bin=self.conn.version < (3, 0)
        )
        res = await self.conn.call16("func_param_bare", [p])
        self.assertDictEqual(res[0][0], cmp, "Body ok")

    async def test__call16_timeout_in_time(self):
        try:
            await self.conn.call16("func_long", [0.1], timeout=1)
        except Exception as e:
            self.fail(e)

    async def test__call_timeout_late(self):
        with self.assertRaises(asyncio.TimeoutError):
            await self.conn.call16("func_long", [0.3], timeout=0.1)



cimport cython


@cython.final
cdef class PrepareRequest(BaseRequest):
    cdef int encode_body(self, WriteBuffer buffer) except -1:
        cdef:
            char *begin
            char *p
            uint32_t body_map_sz
            uint32_t max_body_len

            bytes query_temp
            char *query_str
            ssize_t query_len
            uint32_t kind

        body_map_sz = 1
        max_body_len = 0

        query_str = NULL
        query_len = 0

        if self.query is not None:
            query_temp = encode_unicode_string(self.query, buffer._encoding)
            cpython.bytes.PyBytes_AsStringAndSize(query_temp,
                                                  &query_str,
                                                  &query_len)
            # Size description:
            # mp_sizeof_map()
            # + mp_sizeof_uint(TP_SQL_TEXT)
            # + mp_sizeof_str(query)
            # + mp_sizeof_uint(TP_SQL_BIND)
            max_body_len = 1 \
                           + 1 \
                           + mp_sizeof_str(<uint32_t> query_len) \
                           + 1
            kind = tarantool.IPROTO_SQL_TEXT
        else:
            # Size description:
            # mp_sizeof_map()
            # + mp_sizeof_uint(IPROTO_STMT_ID)
            # + mp_sizeof_int(self.statement_id)
            # + mp_sizeof_uint(TP_SQL_BIND)
            max_body_len = 1 \
                           + 1 \
                           + 9 \
                           + 1
            kind = tarantool.IPROTO_STMT_ID

        buffer.ensure_allocated(max_body_len)

        p = begin = &buffer._buf[buffer._length]
        p = mp_encode_map(p, body_map_sz)
        p = mp_encode_uint(p, kind)
        if query_str != NULL:
            p = mp_encode_str(p, query_str, <uint32_t> query_len)
        else:
            p = mp_encode_uint(p, self.statement_id)

        buffer._length += (p - begin)



from asynctnt import Response
from asynctnt.exceptions import TarantoolSchemaError
from tests import BaseTarantoolTestCase


class DeleteTestCase(BaseTarantoolTestCase):
    async def _fill_data(self):
        data = [
            [0, "a", 1, 2, "hello my darling"],
            [1, "b", 3, 4, "hello my darling, again"],
        ]
        for t in data:
            await self.conn.insert(self.TESTER_SPACE_ID, t)

        return data

    async def test__delete_one(self):
        data = await self._fill_data()

        res = await self.conn.delete(self.TESTER_SPACE_ID, [data[0][0]])
        self.assertIsInstance(res, Response, "Got response")
        self.assertEqual(res.code, 0, "success")
        self.assertGreater(res.sync, 0, "sync > 0")
        self.assertResponseEqual(res, [data[0]], "Body ok")

        res = await self.conn.select(self.TESTER_SPACE_ID, [0])
        self.assertResponseEqual(res, [], "Body ok")

    async def test__delete_by_name(self):
        data = await self._fill_data()

        res = await self.conn.delete(self.TESTER_SPACE_NAME, [data[0][0]])
        self.assertIsInstance(res, Response, "Got response")
        self.assertEqual(res.code, 0, "success")
        self.assertGreater(res.sync, 0, "sync > 0")
        self.assertResponseEqual(res, [data[0]], "Body ok")

        res = await self.conn.select(self.TESTER_SPACE_ID, [0])
        self.assertResponseEqual(res, [], "Body ok")

    async def test__delete_by_index_id(self):
        index_name = "temp_idx"
        res = self.tnt.command('make_third_index("{}")'.format(index_name))
        index_id = res[0][0]

        try:
            await self.tnt_reconnect()

            data = await self._fill_data()

            res = await self.conn.delete(
                self.TESTER_SPACE_NAME, [data[1][2]], index=index_id
            )
            self.assertIsInstance(res, Response, "Got response")
            self.assertEqual(res.code, 0, "success")
            self.assertGreater(res.sync, 0, "sync > 0")
            self.assertResponseEqual(res, [data[1]], "Body ok")

            res = await self.conn.select(
                self.TESTER_SPACE_ID, [data[1][2]], index=index_id
            )
            self.assertResponseEqual(res, [], "Body ok")
        finally:
            self.tnt.command(
                "box.space.{}.index.{}:drop()".format(
                    self.TESTER_SPACE_NAME, index_name
                )
            )

    async def test__delete_by_index_name(self):
        index_name = "temp_idx"
        res = self.tnt.command('make_third_index("{}")'.format(index_name))
        index_id = res[0][0]

        try:
            await self.tnt_reconnect()

            data = await self._fill_data()

            res = await self.conn.delete(
                self.TESTER_SPACE_NAME, [data[1][2]], index=index_name
            )
            self.assertIsInstance(res, Response, "Got response")
            self.assertEqual(res.code, 0, "success")
            self.assertGreater(res.sync, 0, "sync > 0")
            self.assertResponseEqual(res, [data[1]], "Body ok")

            res = await self.conn.select(
                self.TESTER_SPACE_ID, [data[1][2]], index=index_id
            )
            self.assertResponseEqual(res, [], "Body ok")
        finally:
            self.tnt.command(
                "box.space.{}.index.{}:drop()".format(
                    self.TESTER_SPACE_NAME, index_name
                )
            )

    async def test__delete_by_name_no_schema(self):
        await self.tnt_reconnect(fetch_schema=False)

        with self.assertRaises(TarantoolSchemaError):
            await self.conn.delete(self.TESTER_SPACE_NAME, [0])

    async def test__delete_by_index_name_no_schema(self):
        await self.tnt_reconnect(fetch_schema=False)

        with self.assertRaises(TarantoolSchemaError):
            await self.conn.delete(self.TESTER_SPACE_ID, [0], index="primary")

    async def test__delete_invalid_types(self):
        with self.assertRaisesRegex(
            TypeError, "missing 2 required positional arguments: 'space' and 'key'"
        ):
            await self.conn.delete()

    async def test__delete_key_tuple(self):
        try:
            await self.conn.delete(self.TESTER_SPACE_ID, (1,))
        except Exception as e:
            self.fail(e)

    async def test__delete_dict_key(self):
        data = await self._fill_data()

        res = await self.conn.delete(self.TESTER_SPACE_ID, {"f1": 0})
        self.assertResponseEqual(res, [data[0]], "Body ok")

    async def test__delete_dict_resp(self):
        data = [0, "hello", 0, 1, "wow"]
        await self.conn.insert(self.TESTER_SPACE_ID, data)

        res = await self.conn.delete(self.TESTER_SPACE_ID, [0])
        self.assertResponseEqualKV(
            res, [{"f1": 0, "f2": "hello", "f3": 0, "f4": 1, "f5": "wow"}]
        )



cimport asynctnt.iproto.tarantool as tarantool
cimport asynctnt.iproto.tupleobj as tupleobj

include "const.pxi"

include "cmsgpuck.pxd"
include "xd.pxd"
include "python.pxd"
include "bit.pxd"

include "unicodeutil.pxd"
include "schema.pxd"
include "ext.pxd"
include "buffer.pxd"
include "rbuffer.pxd"

include "requests/base.pxd"
include "requests/ping.pxd"
include "requests/call.pxd"
include "requests/eval.pxd"
include "requests/select.pxd"
include "requests/insert.pxd"
include "requests/delete.pxd"
include "requests/update.pxd"
include "requests/upsert.pxd"
include "requests/prepare.pxd"
include "requests/execute.pxd"
include "requests/id.pxd"
include "requests/auth.pxd"
include "requests/streams.pxd"

include "response.pxd"
include "db.pxd"
include "push.pxd"

include "coreproto.pxd"

cdef enum PostConnectionState:
    POST_CONNECTION_NONE = 0
    POST_CONNECTION_ID = 10
    POST_CONNECTION_AUTH = 20
    POST_CONNECTION_SCHEMA = 30
    POST_CONNECTION_DONE = 100


ctypedef object (*req_execute_func)(BaseProtocol, BaseRequest, float)

cdef class BaseProtocol(CoreProtocol):
    cdef:
        object loop
        str username
        str password
        bint fetch_schema
        bint auto_refetch_schema
        float request_timeout
        int post_con_state

        object connected_fut
        object on_connection_made_cb
        object on_connection_lost_cb

        object _on_request_completed_cb
        object _on_request_timeout_cb

        dict _reqs
        uint64_t _sync
        Schema _schema
        int64_t _schema_id
        bint _schema_fetch_in_progress
        object _refetch_schema_future
        Db _db
        req_execute_func execute

        object create_future

    cdef void _set_connection_ready(self)
    cdef void _set_connection_error(self, e)
    cdef void _post_con_state_machine(self)

    cdef void _do_id(self)
    cdef void _do_auth(self, str username, str password)
    cdef void _do_fetch_schema(self, object fut)
    cdef object _refetch_schema(self)

    cdef inline uint64_t next_sync(self)
    cdef inline uint64_t next_stream_id(self)
    cdef uint32_t transform_iterator(self, iterator) except *

    cdef object _new_waiter_for_request(self, Response response, BaseRequest req, float timeout)
    cdef Db _create_db(self, bint gen_stream_id)
    cdef object _execute_bad(self, BaseRequest req, float timeout)
    cdef object _execute_normal(self, BaseRequest req, float timeout)



import enum
from typing import Optional

from asynctnt.iproto.protocol import IProtoError


class TarantoolError(Exception):
    """
    Base Tarantool Exception class
    """

    pass


class TarantoolSchemaError(TarantoolError):
    """
    Exception is raised when any problems with schema occurred
    """

    pass


class TarantoolDatabaseError(TarantoolError):
    """
    Exception is raised when Tarantool responds with code != 0
    """

    def __init__(self, code: int, message: str, error: Optional[IProtoError]):
        super(TarantoolDatabaseError, self).__init__(message)
        self.code = code
        self.message = message
        self.error = error


class TarantoolNetworkError(TarantoolError):
    pass


class TarantoolNotConnectedError(TarantoolNetworkError):
    """
    Raised when asynctnt is not connected to Tarantool
    """

    pass


class ErrorCode(enum.IntEnum):
    """
    Tarantool default error codes
    """

    ER_UNKNOWN = 0
    ER_ILLEGAL_PARAMS = 1
    ER_MEMORY_ISSUE = 2
    ER_TUPLE_FOUND = 3
    ER_TUPLE_NOT_FOUND = 4
    ER_UNSUPPORTED = 5
    ER_NONMASTER = 6
    ER_READONLY = 7
    ER_INJECTION = 8
    ER_CREATE_SPACE = 9
    ER_SPACE_EXISTS = 10
    ER_DROP_SPACE = 11
    ER_ALTER_SPACE = 12
    ER_INDEX_TYPE = 13
    ER_MODIFY_INDEX = 14
    ER_LAST_DROP = 15
    ER_TUPLE_FORMAT_LIMIT = 16
    ER_DROP_PRIMARY_KEY = 17
    ER_KEY_PART_TYPE = 18
    ER_EXACT_MATCH = 19
    ER_INVALID_MSGPACK = 20
    ER_PROC_RET = 21
    ER_TUPLE_NOT_ARRAY = 22
    ER_FIELD_TYPE = 23
    ER_INDEX_PART_TYPE_MISMATCH = 24
    ER_UPDATE_SPLICE = 25
    ER_UPDATE_ARG_TYPE = 26
    ER_FORMAT_MISMATCH_INDEX_PART = 27
    ER_UNKNOWN_UPDATE_OP = 28
    ER_UPDATE_FIELD = 29
    ER_FUNCTION_TX_ACTIVE = 30
    ER_KEY_PART_COUNT = 31
    ER_PROC_LUA = 32
    ER_NO_SUCH_PROC = 33
    ER_NO_SUCH_TRIGGER = 34
    ER_NO_SUCH_INDEX_ID = 35
    ER_NO_SUCH_SPACE = 36
    ER_NO_SUCH_FIELD_NO = 37
    ER_EXACT_FIELD_COUNT = 38
    ER_FIELD_MISSING = 39
    ER_WAL_IO = 40
    ER_MORE_THAN_ONE_TUPLE = 41
    ER_ACCESS_DENIED = 42
    ER_CREATE_USER = 43
    ER_DROP_USER = 44
    ER_NO_SUCH_USER = 45
    ER_USER_EXISTS = 46
    ER_PASSWORD_MISMATCH = 47
    ER_UNKNOWN_REQUEST_TYPE = 48
    ER_UNKNOWN_SCHEMA_OBJECT = 49
    ER_CREATE_FUNCTION = 50
    ER_NO_SUCH_FUNCTION = 51
    ER_FUNCTION_EXISTS = 52
    ER_BEFORE_REPLACE_RET = 53
    ER_MULTISTATEMENT_TRANSACTION = 54
    ER_TRIGGER_EXISTS = 55
    ER_USER_MAX = 56
    ER_NO_SUCH_ENGINE = 57
    ER_RELOAD_CFG = 58
    ER_CFG = 59
    ER_SAVEPOINT_EMPTY_TX = 60
    ER_NO_SUCH_SAVEPOINT = 61
    ER_UNKNOWN_REPLICA = 62
    ER_REPLICASET_UUID_MISMATCH = 63
    ER_INVALID_UUID = 64
    ER_REPLICASET_UUID_IS_RO = 65
    ER_INSTANCE_UUID_MISMATCH = 66
    ER_REPLICA_ID_IS_RESERVED = 67
    ER_INVALID_ORDER = 68
    ER_MISSING_REQUEST_FIELD = 69
    ER_IDENTIFIER = 70
    ER_DROP_FUNCTION = 71
    ER_ITERATOR_TYPE = 72
    ER_REPLICA_MAX = 73
    ER_INVALID_XLOG = 74
    ER_INVALID_XLOG_NAME = 75
    ER_INVALID_XLOG_ORDER = 76
    ER_NO_CONNECTION = 77
    ER_TIMEOUT = 78
    ER_ACTIVE_TRANSACTION = 79
    ER_CURSOR_NO_TRANSACTION = 80
    ER_CROSS_ENGINE_TRANSACTION = 81
    ER_NO_SUCH_ROLE = 82
    ER_ROLE_EXISTS = 83
    ER_CREATE_ROLE = 84
    ER_INDEX_EXISTS = 85
    ER_SESSION_CLOSED = 86
    ER_ROLE_LOOP = 87
    ER_GRANT = 88
    ER_PRIV_GRANTED = 89
    ER_ROLE_GRANTED = 90
    ER_PRIV_NOT_GRANTED = 91
    ER_ROLE_NOT_GRANTED = 92
    ER_MISSING_SNAPSHOT = 93
    ER_CANT_UPDATE_PRIMARY_KEY = 94
    ER_UPDATE_INTEGER_OVERFLOW = 95
    ER_GUEST_USER_PASSWORD = 96
    ER_TRANSACTION_CONFLICT = 97
    ER_UNSUPPORTED_PRIV = 98
    ER_LOAD_FUNCTION = 99
    ER_FUNCTION_LANGUAGE = 100
    ER_RTREE_RECT = 101
    ER_PROC_C = 102
    ER_UNKNOWN_RTREE_INDEX_DISTANCE_TYPE = 103
    ER_PROTOCOL = 104
    ER_UPSERT_UNIQUE_SECONDARY_KEY = 105
    ER_WRONG_INDEX_RECORD = 106
    ER_WRONG_INDEX_PARTS = 107
    ER_WRONG_INDEX_OPTIONS = 108
    ER_WRONG_SCHEMA_VERSION = 109
    ER_MEMTX_MAX_TUPLE_SIZE = 110
    ER_WRONG_SPACE_OPTIONS = 111
    ER_UNSUPPORTED_INDEX_FEATURE = 112
    ER_VIEW_IS_RO = 113
    ER_NO_TRANSACTION = 114
    ER_SYSTEM = 115
    ER_LOADING = 116
    ER_CONNECTION_TO_SELF = 117
    ER_KEY_PART_IS_TOO_LONG = 118
    ER_COMPRESSION = 119
    ER_CHECKPOINT_IN_PROGRESS = 120
    ER_SUB_STMT_MAX = 121
    ER_COMMIT_IN_SUB_STMT = 122
    ER_ROLLBACK_IN_SUB_STMT = 123
    ER_DECOMPRESSION = 124
    ER_INVALID_XLOG_TYPE = 125
    ER_ALREADY_RUNNING = 126
    ER_INDEX_FIELD_COUNT_LIMIT = 127
    ER_LOCAL_INSTANCE_ID_IS_READ_ONLY = 128
    ER_BACKUP_IN_PROGRESS = 129
    ER_READ_VIEW_ABORTED = 130
    ER_INVALID_INDEX_FILE = 131
    ER_INVALID_RUN_FILE = 132
    ER_INVALID_VYLOG_FILE = 133
    ER_CASCADE_ROLLBACK = 134
    ER_VY_QUOTA_TIMEOUT = 135
    ER_PARTIAL_KEY = 136
    ER_TRUNCATE_SYSTEM_SPACE = 137
    ER_LOAD_MODULE = 138
    ER_VINYL_MAX_TUPLE_SIZE = 139
    ER_WRONG_DD_VERSION = 140
    ER_WRONG_SPACE_FORMAT = 141
    ER_CREATE_SEQUENCE = 142
    ER_ALTER_SEQUENCE = 143
    ER_DROP_SEQUENCE = 144
    ER_NO_SUCH_SEQUENCE = 145
    ER_SEQUENCE_EXISTS = 146
    ER_SEQUENCE_OVERFLOW = 147
    ER_NO_SUCH_INDEX_NAME = 148
    ER_SPACE_FIELD_IS_DUPLICATE = 149
    ER_CANT_CREATE_COLLATION = 150
    ER_WRONG_COLLATION_OPTIONS = 151
    ER_NULLABLE_PRIMARY = 152
    ER_NO_SUCH_FIELD_NAME_IN_SPACE = 153
    ER_TRANSACTION_YIELD = 154
    ER_NO_SUCH_GROUP = 155
    ER_SQL_BIND_VALUE = 156
    ER_SQL_BIND_TYPE = 157
    ER_SQL_BIND_PARAMETER_MAX = 158
    ER_SQL_EXECUTE = 159
    ER_UPDATE_DECIMAL_OVERFLOW = 160
    ER_SQL_BIND_NOT_FOUND = 161
    ER_ACTION_MISMATCH = 162
    ER_VIEW_MISSING_SQL = 163
    ER_FOREIGN_KEY_CONSTRAINT = 164
    ER_NO_SUCH_MODULE = 165
    ER_NO_SUCH_COLLATION = 166
    ER_CREATE_FK_CONSTRAINT = 167
    ER_DROP_FK_CONSTRAINT = 168
    ER_NO_SUCH_CONSTRAINT = 169
    ER_CONSTRAINT_EXISTS = 170
    ER_SQL_TYPE_MISMATCH = 171
    ER_ROWID_OVERFLOW = 172
    ER_DROP_COLLATION = 173
    ER_ILLEGAL_COLLATION_MIX = 174
    ER_SQL_NO_SUCH_PRAGMA = 175
    ER_SQL_CANT_RESOLVE_FIELD = 176
    ER_INDEX_EXISTS_IN_SPACE = 177
    ER_INCONSISTENT_TYPES = 178
    ER_SQL_SYNTAX_WITH_POS = 179
    ER_SQL_STACK_OVERFLOW = 180
    ER_SQL_SELECT_WILDCARD = 181
    ER_SQL_STATEMENT_EMPTY = 182
    ER_SQL_KEYWORD_IS_RESERVED = 183
    ER_SQL_SYNTAX_NEAR_TOKEN = 184
    ER_SQL_UNKNOWN_TOKEN = 185
    ER_SQL_PARSER_GENERIC = 186
    ER_SQL_ANALYZE_ARGUMENT = 187
    ER_SQL_COLUMN_COUNT_MAX = 188
    ER_HEX_LITERAL_MAX = 189
    ER_INT_LITERAL_MAX = 190
    ER_SQL_PARSER_LIMIT = 191
    ER_INDEX_DEF_UNSUPPORTED = 192
    ER_CK_DEF_UNSUPPORTED = 193
    ER_MULTIKEY_INDEX_MISMATCH = 194
    ER_CREATE_CK_CONSTRAINT = 195
    ER_CK_CONSTRAINT_FAILED = 196
    ER_SQL_COLUMN_COUNT = 197
    ER_FUNC_INDEX_FUNC = 198
    ER_FUNC_INDEX_FORMAT = 199
    ER_FUNC_INDEX_PARTS = 200
    ER_NO_SUCH_FIELD_NAME = 201
    ER_FUNC_WRONG_ARG_COUNT = 202
    ER_BOOTSTRAP_READONLY = 203
    ER_SQL_FUNC_WRONG_RET_COUNT = 204
    ER_FUNC_INVALID_RETURN_TYPE = 205
    ER_SQL_PARSER_GENERIC_WITH_POS = 206
    ER_REPLICA_NOT_ANON = 207
    ER_CANNOT_REGISTER = 208
    ER_SESSION_SETTING_INVALID_VALUE = 209
    ER_SQL_PREPARE = 210
    ER_WRONG_QUERY_ID = 211
    ER_SEQUENCE_NOT_STARTED = 212
    ER_NO_SUCH_SESSION_SETTING = 213
    ER_UNCOMMITTED_FOREIGN_SYNC_TXNS = 214
    ER_SYNC_MASTER_MISMATCH = 215
    ER_SYNC_QUORUM_TIMEOUT = 216
    ER_SYNC_ROLLBACK = 217
    ER_TUPLE_METADATA_IS_TOO_BIG = 218
    ER_XLOG_GAP = 219
    ER_TOO_EARLY_SUBSCRIBE = 220
    ER_SQL_CANT_ADD_AUTOINC = 221
    ER_QUORUM_WAIT = 222
    ER_INTERFERING_PROMOTE = 223
    ER_ELECTION_DISABLED = 224
    ER_TXN_ROLLBACK = 225
    ER_NOT_LEADER = 226
    ER_SYNC_QUEUE_UNCLAIMED = 227
    ER_SYNC_QUEUE_FOREIGN = 228
    ER_UNABLE_TO_PROCESS_IN_STREAM = 229
    ER_UNABLE_TO_PROCESS_OUT_OF_STREAM = 230
    ER_TRANSACTION_TIMEOUT = 231
    ER_ACTIVE_TIMER = 232
    ER_TUPLE_FIELD_COUNT_LIMIT = 233
    ER_CREATE_CONSTRAINT = 234
    ER_FIELD_CONSTRAINT_FAILED = 235
    ER_TUPLE_CONSTRAINT_FAILED = 236
    ER_CREATE_FOREIGN_KEY = 237
    ER_FOREIGN_KEY_INTEGRITY = 238
    ER_FIELD_FOREIGN_KEY_FAILED = 239
    ER_COMPLEX_FOREIGN_KEY_FAILED = 240
    ER_WRONG_SPACE_UPGRADE_OPTIONS = 241
    ER_NO_ELECTION_QUORUM = 242
    ER_SSL = 243
    ER_SPLIT_BRAIN = 244



import logging
import os
import sys
import unittest

from ._testbase import TarantoolTestCase

CURRENT_DIR = os.path.dirname(os.path.abspath(__file__))


class BaseTarantoolTestCase(TarantoolTestCase):
    DO_CONNECT = True
    LOGGING_LEVEL = getattr(logging, os.getenv("LOG", "CRITICAL").upper())
    LOGGING_STREAM = sys.stdout
    TNT_APP_LUA_PATH = os.path.join(CURRENT_DIR, "files", "app.lua")

    TESTER_SPACE_ID = 512
    TESTER_SPACE_NAME = "tester"

    async def truncate(self):
        if self.conn and self.conn.is_connected:
            await self.conn.call("truncate", timeout=5)

    def tearDown(self):
        if hasattr(self, "conn"):
            self.loop.run_until_complete(self.truncate())
        super().tearDown()


def suite():
    loader = unittest.TestLoader()
    return loader.discover(CURRENT_DIR, pattern="test_*.py")


if __name__ == "__main__":
    runner = unittest.TextTestRunner(verbosity=2)
    result = runner.run(suite())
    sys.exit(not result.wasSuccessful())



cimport cython

cimport asynctnt.iproto.tarantool as tarantool


@cython.final
cdef class CallRequest(BaseRequest):
    cdef int encode_body(self, WriteBuffer buffer) except -1:
        cdef:
            char *begin
            char *p
            uint32_t body_map_sz
            uint32_t max_body_len

            bytes func_name_temp
            char *func_name_str
            ssize_t func_name_len

        func_name_str = NULL
        func_name_len = 0

        func_name_temp = encode_unicode_string(self.func_name, buffer._encoding)
        cpython.bytes.PyBytes_AsStringAndSize(func_name_temp,
                                              &func_name_str,
                                              &func_name_len)
        body_map_sz = 2
        # Size description:
        # mp_sizeof_map()
        # + mp_sizeof_uint(TP_FUNCTION)
        # + mp_sizeof_str(func_name)
        # + mp_sizeof_uint(TP_TUPLE)
        max_body_len = 1 \
                       + 1 \
                       + mp_sizeof_str(<uint32_t> func_name_len) \
                       + 1

        buffer.ensure_allocated(max_body_len)

        p = begin = &buffer._buf[buffer._length]
        p = mp_encode_map(p, body_map_sz)
        p = mp_encode_uint(p, tarantool.IPROTO_FUNCTION_NAME)
        p = mp_encode_str(p, func_name_str, <uint32_t> func_name_len)

        p = mp_encode_uint(p, tarantool.IPROTO_TUPLE)
        buffer._length += (p - begin)
        p = encode_key_sequence(buffer, p, self.args, None, False)



from cpython.version cimport PY_VERSION_HEX


cdef extern from "Python.h":
    char *PyByteArray_AS_STRING(object obj)
    int Py_REFCNT(object obj)


cdef extern from "datetime.h":
    """
    /* Backport for Python 2.x */
    #if PY_MAJOR_VERSION < 3
        #ifndef PyDateTime_DELTA_GET_DAYS
            #define PyDateTime_DELTA_GET_DAYS(o) (((PyDateTime_Delta*)o)->days)
        #endif
        #ifndef PyDateTime_DELTA_GET_SECONDS
            #define PyDateTime_DELTA_GET_SECONDS(o) (((PyDateTime_Delta*)o)->seconds)
        #endif
        #ifndef PyDateTime_DELTA_GET_MICROSECONDS
            #define PyDateTime_DELTA_GET_MICROSECONDS(o) (((PyDateTime_Delta*)o)->microseconds)
        #endif
    #endif
    /* Backport for Python < 3.6 */
    #if PY_VERSION_HEX < 0x030600a4
        #ifndef PyDateTime_TIME_GET_FOLD
            #define PyDateTime_TIME_GET_FOLD(o) ((void)(o), 0)
        #endif
        #ifndef PyDateTime_DATE_GET_FOLD
            #define PyDateTime_DATE_GET_FOLD(o) ((void)(o), 0)
        #endif
    #endif
    /* Backport for Python < 3.6 */
    #if PY_VERSION_HEX < 0x030600a4
        #define __Pyx_DateTime_DateTimeWithFold(year, month, day, hour, minute, second, microsecond, tz, fold) \
            ((void)(fold), PyDateTimeAPI->DateTime_FromDateAndTime(year, month, day, hour, minute, second, \
                microsecond, tz, PyDateTimeAPI->DateTimeType))
        #define __Pyx_DateTime_TimeWithFold(hour, minute, second, microsecond, tz, fold) \
            ((void)(fold), PyDateTimeAPI->Time_FromTime(hour, minute, second, microsecond, tz, PyDateTimeAPI->TimeType))
    #else /* For Python 3.6+ so that we can pass tz */
        #define __Pyx_DateTime_DateTimeWithFold(year, month, day, hour, minute, second, microsecond, tz, fold) \
            PyDateTimeAPI->DateTime_FromDateAndTimeAndFold(year, month, day, hour, minute, second, \
                microsecond, tz, fold, PyDateTimeAPI->DateTimeType)
        #define __Pyx_DateTime_TimeWithFold(hour, minute, second, microsecond, tz, fold) \
            PyDateTimeAPI->Time_FromTimeAndFold(hour, minute, second, microsecond, tz, fold, PyDateTimeAPI->TimeType)
    #endif
    /* Backport for Python < 3.7 */
    #if PY_VERSION_HEX < 0x030700b1
        #define __Pyx_TimeZone_UTC NULL
        #define __Pyx_TimeZone_FromOffset(offset) ((void)(offset), (PyObject*)NULL)
        #define __Pyx_TimeZone_FromOffsetAndName(offset, name) ((void)(offset), (void)(name), (PyObject*)NULL)
    #else
        #define __Pyx_TimeZone_UTC PyDateTime_TimeZone_UTC
        #define __Pyx_TimeZone_FromOffset(offset) PyTimeZone_FromOffset(offset)
        #define __Pyx_TimeZone_FromOffsetAndName(offset, name) PyTimeZone_FromOffsetAndName(offset, name)
    #endif
    /* Backport for Python < 3.10 */
    #if PY_VERSION_HEX < 0x030a00a1
        #ifndef PyDateTime_TIME_GET_TZINFO
            #define PyDateTime_TIME_GET_TZINFO(o) \
                ((((PyDateTime_Time*)o)->hastzinfo) ? ((PyDateTime_Time*)o)->tzinfo : Py_None)
        #endif
        #ifndef PyDateTime_DATE_GET_TZINFO
            #define PyDateTime_DATE_GET_TZINFO(o) \
                ((((PyDateTime_DateTime*)o)->hastzinfo) ? ((PyDateTime_DateTime*)o)->tzinfo : Py_None)
        #endif
    #endif
    """

    # The above macros is Python 3.7+ so we use these instead
    object __Pyx_TimeZone_FromOffset(object offset)


cdef inline object timezone_new(object offset):
    if PY_VERSION_HEX < 0x030700b1:
        from datetime import timezone
        return timezone(offset)
    return __Pyx_TimeZone_FromOffset(offset)



# ruff: noqa: F401

from .connection import Connection, connect
from .iproto.protocol import (
    Db,
    Field,
    IProtoError,
    IProtoErrorStackFrame,
    Iterator,
    Metadata,
    PushIterator,
    Response,
    Schema,
    SchemaIndex,
    SchemaSpace,
    TarantoolTuple,
)

__version__ = "2.2.0"



import asyncio
import contextlib
import functools
import inspect
import logging
import os
import sys
import time
import unittest
from typing import Optional, Tuple

import asynctnt
from asynctnt import TarantoolTuple
from asynctnt.instance import TarantoolSyncDockerInstance, TarantoolSyncInstance

__all__ = (
    "TestCase",
    "TarantoolTestCase",
    "ensure_version",
    "check_version",
    "ensure_bin_version",
)


@contextlib.contextmanager
def silence_asyncio_long_exec_warning():
    def flt(log_record):
        msg = log_record.getMessage()
        return not msg.startswith("Executing ")

    logger = logging.getLogger("asyncio")
    logger.addFilter(flt)
    try:
        yield
    finally:
        logger.removeFilter(flt)


class TestCaseMeta(type(unittest.TestCase)):
    @staticmethod
    def _iter_methods(bases, ns):
        for base in bases:
            for methname in dir(base):
                if not methname.startswith("test_"):
                    continue

                meth = getattr(base, methname)
                if not inspect.iscoroutinefunction(meth):
                    continue

                yield methname, meth

        for methname, meth in ns.items():
            if not methname.startswith("test_"):
                continue

            if not inspect.iscoroutinefunction(meth):
                continue

            yield methname, meth

    def __new__(mcls, name, bases, ns):
        for methname, meth in mcls._iter_methods(bases, ns):

            @functools.wraps(meth)
            def wrapper(self, *args, __meth__=meth, **kwargs):
                self.loop.run_until_complete(__meth__(self, *args, **kwargs))

            ns[methname] = wrapper

        return super().__new__(mcls, name, bases, ns)


class TestCase(unittest.TestCase, metaclass=TestCaseMeta):
    _loop = None
    _use_uvloop = False

    @classmethod
    def setUpClass(cls):
        cls._use_uvloop = os.environ.get("USE_UVLOOP")

        if cls._use_uvloop:
            import uvloop

            uvloop.install()

        try:
            loop = asyncio.get_running_loop()
            # if there is a running loop - close it
            loop.close()
        except RuntimeError:
            pass

        cls._loop = asyncio.new_event_loop()
        asyncio.set_event_loop(cls._loop)

        if cls._use_uvloop:
            import uvloop

            assert isinstance(cls._loop, uvloop.Loop)

    @classmethod
    def tearDownClass(cls):
        if cls._loop is not None:
            cls._loop.close()
            cls._loop = None

    @property
    def loop(self):
        return self._loop

    @contextlib.contextmanager
    def assertRunUnder(self, delta):
        st = time.monotonic()
        try:
            yield
        finally:
            if time.monotonic() - st > delta:
                raise AssertionError("running block took longer than {}".format(delta))

    @classmethod
    def ensure_future(cls, coro_or_future):
        return asyncio.ensure_future(coro_or_future)

    @classmethod
    def sleep(cls, delay, result=None):
        return asyncio.sleep(delay, result)


class TarantoolTestCase(TestCase):
    DO_CONNECT = True
    LOGGING_LEVEL = logging.WARNING
    LOGGING_STREAM = sys.stderr
    TNT_APP_LUA_PATH = None
    TNT_CLEANUP = True
    EXTRA_BOX_CFG = ""

    tnt = None
    in_docker = False

    @classmethod
    def read_applua(cls):
        if cls.TNT_APP_LUA_PATH:
            with open(cls.TNT_APP_LUA_PATH, "r") as f:
                return f.read()

    @classmethod
    def init_instance(cls):
        tnt, in_docker = cls._make_instance(extra_box_cfg=cls.EXTRA_BOX_CFG)
        cls.tnt = tnt
        cls.in_docker = in_docker

    @classmethod
    def setUpClass(cls):
        TestCase.setUpClass()
        logging.basicConfig(level=cls.LOGGING_LEVEL, stream=cls.LOGGING_STREAM)
        if cls.tnt is None:
            tnt, in_docker = cls._make_instance(extra_box_cfg=cls.EXTRA_BOX_CFG)
            cls.tnt = tnt
            cls.in_docker = in_docker

        cls.tnt.start()

    @classmethod
    def make_instance(cls):
        obj, _ = cls._make_instance()
        return obj

    @classmethod
    def _make_instance(cls, **kwargs):
        tarantool_docker_image = os.getenv("TARANTOOL_DOCKER_IMAGE")
        tarantool_docker_tag = os.getenv("TARANTOOL_DOCKER_VERSION")
        in_docker = False
        if tarantool_docker_tag:
            print(
                "Running tarantool in docker: {}:{}".format(
                    tarantool_docker_image or "tarantool/tarantool",
                    tarantool_docker_tag,
                )
            )

            tnt = TarantoolSyncDockerInstance(
                applua=cls.read_applua(),
                docker_image=tarantool_docker_image,
                docker_tag=tarantool_docker_tag,
                timeout=4 * 60,
                **kwargs,
            )
            in_docker = True
        else:
            unix_path = os.getenv("TARANTOOL_LISTEN_UNIX_PATH")
            if not unix_path:
                tnt = TarantoolSyncInstance(
                    port=TarantoolSyncInstance.get_random_port(),
                    console_port=TarantoolSyncInstance.get_random_port(),
                    applua=cls.read_applua(),
                    cleanup=cls.TNT_CLEANUP,
                    **kwargs,
                )
            else:
                tnt = TarantoolSyncInstance(
                    host="unix/",
                    port=unix_path,
                    console_host="127.0.0.1",
                    applua=cls.read_applua(),
                    cleanup=cls.TNT_CLEANUP,
                    **kwargs,
                )

        return tnt, in_docker

    @classmethod
    def tearDownClass(cls):
        if cls.tnt:
            cls.tnt.stop()
        TestCase.tearDownClass()

    def setUp(self):
        super(TarantoolTestCase, self).setUp()
        if self.DO_CONNECT:
            self.loop.run_until_complete(self.tnt_connect())

    def tearDown(self):
        self.loop.run_until_complete(self.tnt_disconnect())
        super(TarantoolTestCase, self).tearDown()

    @property
    def conn(self) -> asynctnt.Connection:
        return self._conn

    async def tnt_connect(
        self,
        *,
        username=None,
        password=None,
        fetch_schema=True,
        auto_refetch_schema=False,
        connect_timeout=None,
        reconnect_timeout=1 / 3,
        ping_timeout=0,
        request_timeout=None,
        encoding="utf-8",
        initial_read_buffer_size=None,
    ):
        self._conn = asynctnt.Connection(
            host=self.tnt.host,
            port=self.tnt.port,
            username=username,
            password=password,
            fetch_schema=fetch_schema,
            auto_refetch_schema=auto_refetch_schema,
            connect_timeout=connect_timeout,
            reconnect_timeout=reconnect_timeout,
            request_timeout=request_timeout,
            ping_timeout=ping_timeout,
            encoding=encoding,
            initial_read_buffer_size=initial_read_buffer_size,
        )
        await self._conn.connect()
        return self._conn

    async def tnt_disconnect(self):
        if hasattr(self, "conn") and self.conn is not None:
            await self._conn.disconnect()
            self._conn = None

    async def tnt_reconnect(self, **kwargs):
        await self.tnt_disconnect()
        await self.tnt_connect(**kwargs)

    def assertResponseEqual(self, resp, target, *args):
        tuples = []
        for item in resp:
            if isinstance(item, TarantoolTuple):
                item = list(item)
            tuples.append(item)
        return self.assertListEqual(tuples, target, *args)

    def assertResponseEqualKV(self, resp, target, *args):
        tuples = []
        for item in resp:
            if isinstance(item, TarantoolTuple):
                item = dict(item)
            tuples.append(item)
        return self.assertListEqual(tuples, target, *args)


def ensure_version(*, min=None, max=None, min_included=True, max_included=False):
    def check_version_wrap(f):
        @functools.wraps(f)
        async def wrap(self, *args, **kwargs):
            if check_version(
                self,
                self.conn.version,
                min=min,
                max=max,
                min_included=min_included,
                max_included=max_included,
            ):
                res = f(self, *args, **kwargs)
                if inspect.isawaitable(res):
                    return await res
                return res

        return wrap

    return check_version_wrap


def ensure_bin_version(*, min=None, max=None, min_included=True, max_included=False):
    def check_version_wrap(cls):
        assert issubclass(cls, TarantoolTestCase)
        cls.init_instance()
        ok, reason = _check_version(
            cls.tnt.bin_version,
            min=min,
            max=max,
            min_included=min_included,
            max_included=max_included,
        )

        if not ok:
            cls = unittest.skip(reason)(cls)

        return cls

    return check_version_wrap


def check_version(
    test, version, *, min=None, max=None, min_included=False, max_included=False
):
    ok, reason = _check_version(
        version, min=min, max=max, min_included=min_included, max_included=max_included
    )

    if not ok:
        test.skipTest(reason)
        return False

    return True


def _check_version(
    version, *, min=None, max=None, min_included=False, max_included=False
) -> Tuple[bool, Optional[str]]:
    if min and (version < min or (min_included and version <= min)):
        return False, f"version mismatch - required min={min} got={version}"

    if max and (version > max or (max_included and version >= max)):
        return False, f"version mismatch - required max={max} got={version}"

    return True, None



from libc.stdint cimport int64_t


cdef class Field:
    cdef:
        readonly str name
        readonly str type
        readonly str collation
        readonly object is_nullable
        readonly object is_autoincrement
        readonly str span


cdef public class Metadata [object C_Metadata, type C_Metadata_Type]:
    cdef:
        readonly list fields
        readonly dict name_id_map
        list names

    cdef inline int len(self)
    cdef inline void add(self, int id, Field field)
    cdef inline str name_by_id(self, int i)
    cdef inline int id_by_name(self, str name) except *
    cdef inline int id_by_name_safe(self, str name) except*


cdef class SchemaIndex:
    cdef:
        readonly int sid
        readonly int iid
        readonly str name
        readonly str index_type
        readonly object unique
        readonly Metadata metadata


cdef class SchemaDummyIndex(SchemaIndex):
    pass


cdef class SchemaSpace:
    cdef:
        readonly int sid
        readonly int owner
        readonly str name
        readonly str engine
        readonly int field_count
        readonly object flags

        readonly Metadata metadata
        readonly dict indexes

    cdef void add_index(self, SchemaIndex idx)
    cdef SchemaIndex get_index(self, index, create_dummy=*)


cdef class SchemaDummySpace(SchemaSpace):
    pass


cdef class Schema:
    cdef:
        readonly dict spaces
        readonly int id

    cdef SchemaSpace get_space(self, space)
    cdef SchemaSpace create_dummy_space(self, int space_id)
    cdef SchemaSpace get_or_create_space(self, space)

    cdef SchemaSpace parse_space(self, space_row)
    cdef SchemaIndex parse_index(self, index_row)

    cdef inline clear(self)

    @staticmethod
    cdef Schema parse(int64_t schema_id, spaces, indexes)


cdef list dict_to_list_fields(dict d, Metadata metadata, bint default_none)



from cpython.datetime cimport datetime
from libc cimport math
from libc.stdint cimport int16_t, int32_t, int64_t, uint8_t, uint32_t


cdef inline uint32_t bcd_len(uint32_t digits_len):
    return <uint32_t> math.floor(digits_len / 2) + 1

cdef uint32_t decimal_len(int exponent, uint32_t digits_count)
cdef char *decimal_encode(char *p,
                          uint32_t digits_count,
                          uint8_t sign,
                          tuple digits,
                          int exponent) except NULL
cdef object decimal_decode(const char ** p, uint32_t length)

cdef object uuid_decode(const char ** p, uint32_t length)

cdef struct IProtoDateTime:
    int64_t seconds
    int32_t nsec
    int16_t tzoffset
    int16_t tzindex

cdef void datetime_zero(IProtoDateTime *dt)
cdef uint32_t datetime_len(IProtoDateTime *dt)
cdef char *datetime_encode(char *p, IProtoDateTime *dt) except NULL
cdef int datetime_decode(const char ** p,
                         uint32_t length,
                         IProtoDateTime *dt) except -1
cdef void datetime_from_py(datetime ob, IProtoDateTime *dt)
cdef object datetime_to_py(IProtoDateTime *dt)



from asynctnt import Response
from asynctnt.prepared import PreparedStatement
from tests import BaseTarantoolTestCase
from tests._testbase import ensure_version


class SQLPreparedStatementTestCase(BaseTarantoolTestCase):
    @ensure_version(min=(2, 0))
    async def test__basic(self):
        stmt = self.conn.prepare("select 1, 2")
        self.assertIsInstance(stmt, PreparedStatement, "Got correct instance")
        async with stmt:
            self.assertIsNotNone(stmt.id, "statement has been prepared")

            res = await stmt.execute()
            self.assertIsInstance(res, Response, "Got response")
            self.assertEqual(res.code, 0, "success")
            self.assertGreater(res.sync, 0, "sync > 0")
            self.assertResponseEqual(res, [[1, 2]], "Body ok")

        self.assertIsNone(stmt.id, "statement has been unprepared")

    @ensure_version(min=(2, 0))
    async def test__manual(self):
        stmt = self.conn.prepare("select 1, 2")
        self.assertIsInstance(stmt, PreparedStatement, "Got correct instance")
        stmt_id = await stmt.prepare()
        self.assertIsNotNone(stmt_id, "statement has been prepared")
        res = await stmt.execute()
        self.assertIsInstance(res, Response, "Got response")
        self.assertEqual(res.code, 0, "success")
        self.assertGreater(res.sync, 0, "sync > 0")
        self.assertResponseEqual(res, [[1, 2]], "Body ok")
        await stmt.unprepare()

    @ensure_version(min=(2, 0))
    async def test__manual_iproto(self):
        res = await self.conn.prepare_iproto("select 1, 2")
        self.assertEqual(res.code, 0, "success")
        stmt_id = res.stmt_id
        self.assertNotEqual(stmt_id, 0, "received statement_id")

        res = await self.conn.execute(stmt_id, [])
        self.assertIsInstance(res, Response, "Got response")
        self.assertEqual(res.code, 0, "success")
        self.assertGreater(res.sync, 0, "sync > 0")
        self.assertResponseEqual(res, [[1, 2]], "Body ok")

        res = await self.conn.unprepare_iproto(stmt_id)
        self.assertEqual(res.code, 0, "success")

    @ensure_version(min=(2, 0))
    async def test__bind(self):
        stmt = self.conn.prepare("select 1, 2 where 1 = ? and 2 = ?")
        async with stmt:
            res = await stmt.execute([1, 2])
            self.assertIsInstance(res, Response, "Got response")
            self.assertEqual(res.code, 0, "success")
            self.assertGreater(res.sync, 0, "sync > 0")
            self.assertResponseEqual(res, [[1, 2]], "Body ok")

    @ensure_version(min=(2, 0))
    async def test__bind_metadata(self):
        stmt = self.conn.prepare("select 1, 2 where 1 = :a and 2 = :b")
        async with stmt:
            self.assertIsNotNone(stmt.params_count)
            self.assertIsNotNone(stmt.params)

            self.assertEqual(2, stmt.params_count)
            self.assertEqual(":a", stmt.params.fields[0].name)
            self.assertEqual("ANY", stmt.params.fields[0].type)
            self.assertEqual(":b", stmt.params.fields[1].name)
            self.assertEqual("ANY", stmt.params.fields[1].type)

    @ensure_version(min=(2, 0))
    async def test__bind_2_execute(self):
        stmt = self.conn.prepare("select 1, 2 where 1 = ? and 2 = ?")
        async with stmt:
            res = await stmt.execute([1, 2])
            self.assertIsInstance(res, Response, "Got response")
            self.assertEqual(res.code, 0, "success")
            self.assertGreater(res.sync, 0, "sync > 0")
            self.assertResponseEqual(res, [[1, 2]], "Body ok")

            res = await stmt.execute([3, 4])
            self.assertIsInstance(res, Response, "Got response")
            self.assertEqual(res.code, 0, "success")
            self.assertGreater(res.sync, 0, "sync > 0")
            self.assertResponseEqual(res, [], "Body is empty")

    @ensure_version(min=(2, 0))
    async def test__context_manager_double_enter(self):
        stmt = self.conn.prepare("select 1, 2 where 1 = ? and 2 = ?")
        async with stmt:
            async with stmt:  # does nothing
                res = await stmt.execute([1, 2])
                self.assertResponseEqual(res, [[1, 2]], "Body ok")



import asyncio

cimport cython

import asynctnt


cdef class PushIterator:
    def __init__(self, fut: asyncio.Future):
        """
            Creates PushIterator object. In order to receive push notifications
            this iterator must be created.

            Example:

            .. code-block:: pycon

                # tarantool function:
                # function f()
                #     box.session.push('hello')
                #     return 'finished'
                # end

                >>> fut = conn.call('async_function', push_subscribe=True)
                >>> async for item in PushIterator(fut):
                ...     print(item)


            :param fut: Future object returned from call_async, eval_sync
                        functions
            :type fut: asyncio.Future
        """
        cdef:
            Response response
            BaseRequest request
        if not hasattr(fut, '_response'):
            raise ValueError('Future is invalid. Make sure to call with '
                             'a future returned from a method with '
                             'push_subscribe=True flag')

        response = <Response> fut._response
        request = response.request_

        if not request.push_subscribe:
            raise ValueError('Future is invalid. Make sure to call with '
                             'a future returned from a method with '
                             'push_subscribe=True flag')

        self._fut = fut
        self._request = request
        self._response = response

    def __iter__(self):
        raise RuntimeError('Cannot use iter with PushIterator - use aiter')

    def __next__(self):
        raise RuntimeError('Cannot use next with PushIterator - use anext')

    def __aiter__(self):
        return self

    @cython.iterable_coroutine
    async def __anext__(self):
        cdef Response response
        response = self._response

        if response.push_len() == 0 and response.code_ >= 0:
            # no more data left
            raise StopAsyncIteration

        if response.push_len() > 0:
            return response.pop_push()

        ev = response._push_event
        await ev.wait()

        exc = response.get_exception()
        if exc is not None:
            # someone needs to await the underlying future
            # so we do it here, and most probably (like 99%) the exception
            # that happened is already in the self._fut. So, await-ing it
            # would cause an exception to be thrown.
            # But if it doesnt't throw we still await the future and throw
            # the exception ourselves.
            await self._fut
            raise exc

        if response.push_len() > 0:
            return response.pop_push()

        if response.code_ >= 0:
            ev.clear()
            raise StopAsyncIteration

        assert False, 'Impossible condition happened. ' \
                      'Please file a bug to ' \
                      'https://github.com/igorcoding/asynctnt'

    @property
    def response(self):
        """
            Return current Response object. Might be handy to know if the
            request is finished already while iterating over the PushIterator

            :rtype: asynctnt.Response
        """
        return self._response



cimport cython
from libc.stdint cimport uint32_t


cdef inline size_t size_t_max(size_t a, size_t b):
    if a > b:
        return a
    return b

cdef inline uint32_t nearest_power_of_2(uint32_t v):
    v -= 1
    v |= v >> 1
    v |= v >> 2
    v |= v >> 4
    v |= v >> 8
    v |= v >> 16
    v += 1
    return v

@cython.final
cdef class ReadBuffer:
    cdef:
        char *buf
        size_t initial_buffer_size  # Initial buffer size, obviously
        size_t len  # Allocated size
        size_t use  # Used size

        str encoding

    @staticmethod
    cdef ReadBuffer create(str encoding, size_t initial_buffer_size= *)

    cdef void _reallocate(self, size_t new_size) except *
    cdef int extend(self, const char *data, size_t len) except -1
    cdef void move(self, size_t pos)
    cdef void move_offset(self, ssize_t offset, size_t size) except *
    cdef bytes get_slice(self, size_t begin, size_t end)
    cdef bytes get_slice_begin(self, size_t begin)
    cdef bytes get_slice_end(self, size_t end)



cimport cpython.datetime
from cpython.datetime cimport PyDateTimeAPI, datetime, datetime_tzinfo, timedelta_new
from libc.stdint cimport uint32_t
from libc.string cimport memcpy

from decimal import Decimal
from uuid import UUID


cdef uint32_t decimal_len(int exponent, uint32_t digits_count):
    cdef:
        uint32_t length

    length = bcd_len(digits_count)
    if exponent > 0:
        length += mp_sizeof_int(-exponent)
    else:
        length += mp_sizeof_uint(-exponent)

    return length

cdef char *decimal_encode(char *p,
                          uint32_t digits_count,
                          uint8_t sign,
                          tuple digits,
                          int exponent) except NULL:
    cdef:
        int i
        uint8_t byte
        char *out
        uint32_t length

    # encode exponent
    if exponent > 0:
        p = mp_encode_int(p, -exponent)
    else:
        p = mp_encode_uint(p, -exponent)

    length = bcd_len(digits_count)

    out = &p[length - 1]
    if sign == 1:
        byte = 0x0d
    else:
        byte = 0x0c

    i = digits_count - 1
    while out >= p:
        if i >= 0:
            byte |= (<uint8_t> <object> cpython.tuple.PyTuple_GET_ITEM(digits, i)) << 4

        out[0] = byte
        byte = 0

        if i > 0:
            byte = <uint8_t> <object> cpython.tuple.PyTuple_GET_ITEM(digits, i - 1) & 0xf

        out -= 1
        i -= 2

    p = &p[length]
    return p

cdef object decimal_decode(const char ** p, uint32_t length):
    cdef:
        int exponent
        uint8_t sign
        mp_type obj_type
        const char *first
        const char *last
        uint32_t digits_count
        uint8_t nibble

    sign = 0
    first = &p[0][0]
    last = first + length - 1

    # decode exponent
    obj_type = mp_typeof(p[0][0])
    if obj_type == MP_UINT:
        exponent = -<int> mp_decode_uint(p)
    elif obj_type == MP_INT:
        exponent = -<int> mp_decode_int(p)
    else:
        raise TypeError('unexpected exponent type: {}'.format(obj_type))

    length -= (&p[0][0] - first)
    first = &p[0][0]

    while first[0] == 0:
        first += 1  # skipping leading zeros

    sign = last[0] & 0xf  # extract sign
    if sign == 0x0a or sign == 0x0c or sign == 0x0e or sign == 0x0f:
        sign = 0
    else:
        sign = 1

    # decode digits
    digits_count = (last - first) * 2 + 1
    if first[0] & 0xf0 == 0:
        digits_count -= 1  # adjust for leading zero nibble

    digits = cpython.tuple.PyTuple_New(digits_count)

    if digits_count > 0:
        while True:
            nibble = (last[0] & 0xf0) >> 4  # left nibble first
            item = <object> <int> nibble
            cpython.Py_INCREF(item)
            cpython.tuple.PyTuple_SET_ITEM(digits, digits_count - 1, item)

            digits_count -= 1
            if digits_count == 0:
                break
            last -= 1

            nibble = last[0] & 0x0f  # right nibble
            item = <object> <int> nibble
            cpython.Py_INCREF(item)
            cpython.tuple.PyTuple_SET_ITEM(digits, digits_count - 1, item)

            digits_count -= 1
            if digits_count == 0:
                break

    p[0] += length

    return Decimal((<object> <int> sign, digits, <object> exponent))

cdef object uuid_decode(const char ** p, uint32_t length):
    data = cpython.bytes.PyBytes_FromStringAndSize(p[0], length)
    p[0] += length
    return UUID(bytes=data)

cdef inline void datetime_zero(IProtoDateTime *dt):
    dt.seconds = 0
    dt.nsec = 0
    dt.tzoffset = 0
    dt.tzindex = 0

cdef inline uint32_t datetime_len(IProtoDateTime *dt):
    cdef uint32_t sz
    sz = sizeof(int64_t)
    if dt.nsec != 0 or dt.tzoffset != 0 or dt.tzindex != 0:
        return sz + DATETIME_TAIL_SZ
    return sz

cdef char *datetime_encode(char *p, IProtoDateTime *dt) except NULL:
    store_u64(p, dt.seconds)
    p += sizeof(dt.seconds)
    if dt.nsec != 0 or dt.tzoffset != 0 or dt.tzindex != 0:
        memcpy(p, &dt.nsec, DATETIME_TAIL_SZ)
        p += DATETIME_TAIL_SZ
    return p

cdef int datetime_decode(
        const char ** p,
        uint32_t length,
        IProtoDateTime *dt
) except -1:
    delta = None
    tz = None

    dt.seconds = load_u64(p[0])
    p[0] += sizeof(dt.seconds)
    length -= sizeof(dt.seconds)

    if length == 0:
        return 0

    if length != DATETIME_TAIL_SZ:
        raise ValueError("invalid datetime size. got {} extra bytes".format(
            length
        ))

    dt.nsec = load_u32(p[0])
    p[0] += 4
    dt.tzoffset = load_u16(p[0])
    p[0] += 2
    dt.tzindex = load_u16(p[0])
    p[0] += 2

cdef void datetime_from_py(datetime ob, IProtoDateTime *dt):
    cdef:
        double ts
        int offset
    ts = <double> ob.timestamp()
    dt.seconds = <int64_t> ts
    dt.nsec = <int32_t> ((ts - <double> dt.seconds) * 1000000) * 1000
    if dt.nsec < 0:
        # correction for negative dates
        dt.seconds -= 1
        dt.nsec += 1000000000

    if datetime_tzinfo(ob) is not None:
        offset = ob.utcoffset().total_seconds()
        dt.tzoffset = <int16_t> (offset / 60)

cdef object datetime_to_py(IProtoDateTime *dt):
    cdef:
        double timestamp
        object tz

    tz = None

    if dt.tzoffset != 0:
        delta = timedelta_new(0, <int> dt.tzoffset * 60, 0)
        tz = timezone_new(delta)

    timestamp = dt.seconds + (<double> dt.nsec) / 1e9
    return PyDateTimeAPI.DateTime_FromTimestamp(
        <PyObject *>PyDateTimeAPI.DateTimeType,
        (timestamp,) if tz is None else (timestamp, tz),
        NULL,
    )



cimport cython
from libc.stdint cimport int64_t, uint32_t, uint64_t


@cython.final
cdef class WriteBuffer:
    cdef:
        # Preallocated small buffer
        bint _smallbuf_inuse
        char _smallbuf[_BUFFER_INITIAL_SIZE]

        char *_buf
        ssize_t _size  # Allocated size
        ssize_t _length  # Length of data in the buffer
        int _view_count  # Number of memoryviews attached to the buffer

        bytes _encoding

    @staticmethod
    cdef WriteBuffer create(bytes encoding)

    cdef inline _check_readonly(self)
    cdef inline len(self)
    cdef int ensure_allocated(self, ssize_t extra_length) except -1
    cdef char *_ensure_allocated(self, char *p,
                                 ssize_t extra_length) except NULL
    cdef int _reallocate(self, ssize_t new_size) except -1
    cdef int write_buffer(self, WriteBuffer buf) except -1
    cdef int write_header(self, uint64_t sync,
                          tarantool.iproto_type op,
                          int64_t schema_id,
                          uint64_t stream_id) except -1
    cdef void write_length(self)

    cdef char *mp_encode_nil(self, char *p) except NULL
    cdef char *mp_encode_bool(self, char *p, bint value) except NULL
    cdef char *mp_encode_double(self, char *p, double value) except NULL
    cdef char *mp_encode_uint(self, char *p, uint64_t value) except NULL
    cdef char *mp_encode_int(self, char *p, int64_t value) except NULL
    cdef char *mp_encode_str(self, char *p,
                             const char *str, uint32_t len) except NULL
    cdef char *mp_encode_bin(self, char *p,
                             const char *data, uint32_t len) except NULL
    cdef char *mp_encode_decimal(self, char *p, object value) except NULL
    cdef char *mp_encode_uuid(self, char *p, object value) except NULL
    cdef char *mp_encode_datetime(self, char *p, object value) except NULL
    cdef char *mp_encode_array(self, char *p, uint32_t len) except NULL
    cdef char *mp_encode_map(self, char *p, uint32_t len) except NULL
    cdef char *mp_encode_list(self, char *p, list arr) except NULL
    cdef char *mp_encode_tuple(self, char *p, tuple t) except NULL
    cdef char *mp_encode_dict(self, char *p, dict d) except NULL
    cdef char *mp_encode_obj(self, char *p, object o) except NULL



import asyncio

import asynctnt
from asynctnt import PushIterator, Response
from asynctnt.exceptions import TarantoolDatabaseError, TarantoolNotConnectedError
from tests import BaseTarantoolTestCase
from tests._testbase import ensure_version


class PushTestCase(BaseTarantoolTestCase):
    @ensure_version(min=(1, 10))
    async def test__push_invalid_future(self):
        with self.assertRaises(ValueError) as e:
            PushIterator(asyncio.Future())

        self.assertEqual(
            str(e.exception),
            "Future is invalid. Make sure to call with "
            "a future returned from a method with "
            "push_subscribe=True flag",
        )

    @ensure_version(min=(1, 10))
    async def test__push_invalid_future_no_flag(self):
        res = self.conn.call("async_action")

        with self.assertRaises(ValueError) as e:
            PushIterator(res)

        self.assertEqual(
            str(e.exception),
            "Future is invalid. Make sure to call with "
            "a future returned from a method with "
            "push_subscribe=True flag",
        )

    @ensure_version(min=(1, 10))
    async def test__push_correct_res(self):
        fut = self.conn.call("async_action", push_subscribe=True)
        self.assertEqual(type(fut), asyncio.Future)

        try:
            it = PushIterator(fut)
            self.assertIsInstance(it.response, asynctnt.Response)
        except Exception as e:
            self.fail(e)

    @ensure_version(min=(1, 10))
    async def test__push_call_iter(self):
        fut = self.conn.call("async_action", push_subscribe=True)

        with self.assertRaises(RuntimeError) as e:
            for _ in PushIterator(fut):
                pass
        self.assertEqual(
            str(e.exception), "Cannot use iter with PushIterator - use aiter"
        )

    @ensure_version(min=(1, 10))
    async def test__push_read_all(self):
        fut = self.conn.call("async_action", push_subscribe=True)
        it = PushIterator(fut)

        self.assertFalse(it.response.done(), "response not done")

        result = []

        async for entry in it:
            result.append(entry[0])

        self.assertTrue(it.response.done(), "response is done")

        self.assertListEqual(
            result,
            ["hello_1", "hello_2", "hello_3", "hello_4", "hello_5"],
            "push values ok",
        )

        fut_res = await fut
        self.assertIsInstance(fut_res, Response, "got response")
        self.assertEqual(fut_res.code, 0, "code ok")
        self.assertEqual(fut_res.sync, it.response.sync, "sync ok")
        self.assertEqual(fut_res.return_code, 0, "return code ok")
        self.assertEqual(fut_res.body, ["ret"], "return value ok")
        self.assertTrue(fut_res.done(), "response done")

    @ensure_version(min=(1, 10))
    async def test__push_read_in_parts(self):
        fut = self.conn.call("async_action", push_subscribe=True)
        it = PushIterator(fut)

        result = []

        i = 0
        async for entry in it:
            if len(entry) == 0:
                self.fail("got 0 length for entry #{}".format(i))
            result.append(entry[0])
            i += 1
            if i == 2:
                break

        async for entry in it:
            if len(entry) == 0:
                self.fail("got 0 length for entry #{}".format(i))
            result.append(entry[0])
            i += 1

        self.assertListEqual(
            result,
            ["hello_1", "hello_2", "hello_3", "hello_4", "hello_5"],
            "push values ok",
        )

        fut_res = await fut
        self.assertIsInstance(fut_res, Response, "got response")
        self.assertEqual(fut_res.code, 0, "code ok")
        self.assertEqual(fut_res.sync, it.response.sync, "sync ok")
        self.assertEqual(fut_res.return_code, 0, "return code ok")
        self.assertEqual(fut_res.body, ["ret"], "return value ok")

    @ensure_version(min=(1, 10))
    async def test__push_read_all_eval(self):
        fut = self.conn.eval(
            """
            for i = 1, 5 do
                box.session.push('hello_' .. tostring(i))
                require'fiber'.sleep(0.01)
            end
            return 'ret'
        """,
            push_subscribe=True,
        )
        it = PushIterator(fut)

        result = []

        i = 0
        async for entry in it:
            if len(entry) == 0:
                self.fail("got 0 length for entry #{}".format(i))
            result.append(entry[0])
            i += 1

        self.assertListEqual(
            result,
            ["hello_1", "hello_2", "hello_3", "hello_4", "hello_5"],
            "push values ok",
        )

        fut_res = await fut
        self.assertIsInstance(fut_res, Response, "got response")
        self.assertEqual(fut_res.code, 0, "code ok")
        self.assertEqual(fut_res.sync, it.response.sync, "sync ok")
        self.assertEqual(fut_res.return_code, 0, "return code ok")
        self.assertEqual(fut_res.body, ["ret"], "return value ok")

    @ensure_version(min=(1, 10))
    async def test__push_read_all_various_sleep(self):
        fut = self.conn.eval(
            """
            box.session.push('hello_1')
            require'fiber'.sleep(0.01)
            box.session.push('hello_2')
            require'fiber'.sleep(1)
            box.session.push('hello_3')
            return 'ret'
        """,
            push_subscribe=True,
        )
        it = PushIterator(fut)

        result = []

        i = 0
        async for entry in it:
            result.append(entry[0])
            i += 1

        self.assertListEqual(
            result, ["hello_1", "hello_2", "hello_3"], "push values ok"
        )

        fut_res = await fut
        self.assertIsInstance(fut_res, Response, "got response")
        self.assertEqual(fut_res.code, 0, "code ok")
        self.assertEqual(fut_res.sync, it.response.sync, "sync ok")
        self.assertEqual(fut_res.return_code, 0, "return code ok")
        self.assertEqual(fut_res.body, ["ret"], "return value ok")

    @ensure_version(min=(1, 10))
    async def test__push_read_all_error(self):
        fut = self.conn.eval(
            """
                    for i = 1, 5 do
                        box.session.push('hello_' .. tostring(i))
                        require'fiber'.sleep(0.01)
                    end
                    return 'ret'
                """,
            push_subscribe=True,
        )
        it = PushIterator(fut)

        # iter once
        await it.__anext__()

        # drop tarantool
        self.tnt.stop()

        try:
            with self.assertRaises(TarantoolNotConnectedError):
                await asyncio.wait_for(it.__anext__(), timeout=5)
        finally:
            self.tnt.start()

    @ensure_version(min=(1, 10))
    async def test__push_read_all_disconnect(self):
        fut = self.conn.eval("error('some error')", push_subscribe=True)
        it = PushIterator(fut)

        with self.assertRaises(TarantoolDatabaseError):
            await it.__anext__()

        with self.assertRaises(TarantoolDatabaseError):
            await fut

    @ensure_version(min=(1, 10))
    async def test__push_read_all_multiple_iterators(self):
        fut = self.conn.eval(
            "box.session.push(1);" "box.session.push(2);" "box.session.push(3);",
            push_subscribe=True,
        )
        it1 = PushIterator(fut)
        it2 = PushIterator(fut)

        async def f(it):
            results = []
            async for entry in it:
                results.append(entry[0])

            return results

        res1, res2 = await asyncio.gather(f(it1), f(it2))
        res1.extend(res2)
        res1.sort()

        self.assertListEqual(res1, [1, 2, 3])

    @ensure_version(min=(1, 10))
    async def test__push_read_all_one_iterator(self):
        fut = self.conn.eval(
            "box.session.push('hello_1');"
            "box.session.push('hello_2');"
            "box.session.push('hello_3');",
            push_subscribe=True,
        )

        it = PushIterator(fut)
        results = []
        async for entry in it:
            results.append(entry[0])

        self.assertListEqual(
            results,
            [
                "hello_1",
                "hello_2",
                "hello_3",
            ],
            "push ok",
        )



cimport cython

DEF IPROTO_VERSION = 3

@cython.final
cdef class IDRequest(BaseRequest):
    cdef int encode_body(self, WriteBuffer buffer) except -1:
        cdef:
            char *p
            char *begin
            uint32_t body_map_sz
            uint32_t max_body_len

        body_map_sz = 2
        # Size description:
        # mp_sizeof_map(body_map_sz)
        # + mp_sizeof_uint(tarantool.IPROTO_VERSION)
        # + mp_sizeof_uint(IPROTO_VERSION)
        # + mp_sizeof_uint(tarantool.IPROTO_FEATURES)
        # + mp_sizeof_array(0)
        # + max arr size
        max_body_len = 1 \
                       + 1 \
                       + 1 \
                       + 1 \
                       + 1 \
                       + 4  # Note: maximum 4 elements in array

        buffer.ensure_allocated(max_body_len)

        p = begin = &buffer._buf[buffer._length]
        p = mp_encode_map(p, body_map_sz)
        p = mp_encode_uint(p, tarantool.IPROTO_VERSION)
        p = mp_encode_uint(p, IPROTO_VERSION)
        p = mp_encode_uint(p, tarantool.IPROTO_FEATURES)
        p = mp_encode_array(p, 3)
        p = mp_encode_uint(p, tarantool.IPROTO_FEATURE_STREAMS)
        p = mp_encode_uint(p, tarantool.IPROTO_FEATURE_TRANSACTIONS)
        p = mp_encode_uint(p, tarantool.IPROTO_FEATURE_ERROR_EXTENSION)

        buffer._length += (p - begin)



cimport cpython
cimport cpython.bytes
cimport cpython.dict
cimport cpython.list
cimport cpython.tuple
cimport cpython.unicode
cimport cython
from cpython.datetime cimport datetime
from cpython.mem cimport PyMem_Free, PyMem_Malloc, PyMem_Realloc
from cpython.ref cimport PyObject
from libc.stdint cimport int64_t, uint8_t, uint32_t, uint64_t
from libc.stdio cimport printf
from libc.string cimport memcpy

from decimal import Decimal  # pragma: nocover
from uuid import UUID  # pragma: nocover


# noinspection PyUnresolvedReferences
# noinspection PyAttributeOutsideInit
@cython.no_gc_clear
@cython.final
@cython.freelist(_BUFFER_FREELIST_SIZE)
cdef class WriteBuffer:
    def __cinit__(self):
        self._smallbuf_inuse = True
        self._buf = self._smallbuf
        self._size = _BUFFER_INITIAL_SIZE
        self._length = 0
        self._encoding = None

    def __dealloc__(self):
        if self._buf is not NULL and not self._smallbuf_inuse:
            PyMem_Free(self._buf)
            self._buf = NULL
            self._size = 0

        if self._view_count:  # pragma: nocover
            raise RuntimeError(
                'Deallocating buffer with attached memoryviews')

    def __getbuffer__(self, Py_buffer *buffer, int flags):
        self._view_count += 1

        cpython.PyBuffer_FillInfo(
            buffer, self, self._buf, self._length,
            1,  # read-only
            flags
        )

    def __releasebuffer__(self, Py_buffer *buffer):
        self._view_count -= 1

    def hex(self):  # pragma: nocover
        return ":".join("{:02x}".format(ord(<bytes> (self._buf[i])))
                        for i in range(self._length))

    @staticmethod
    cdef WriteBuffer create(bytes encoding):
        cdef WriteBuffer buf
        buf = WriteBuffer.__new__(WriteBuffer)
        buf._encoding = encoding
        return buf

    cdef inline _check_readonly(self):  # pragma: nocover
        if self._view_count:
            raise BufferError('the buffer is in read-only mode')

    cdef inline len(self):
        return self._length

    cdef int ensure_allocated(self, ssize_t extra_length) except -1:
        cdef ssize_t new_size = extra_length + self._length

        if new_size > self._size:
            self._reallocate(new_size)

    cdef char *_ensure_allocated(self, char *p,
                                 ssize_t extra_length) except NULL:
        cdef:
            ssize_t new_size

        new_size = extra_length + self._length

        if new_size > self._size:
            used = p - self._buf
            self._reallocate(new_size)
            p = &self._buf[used]
        return p

    cdef int _reallocate(self, ssize_t new_size) except -1:
        cdef char *new_buf

        if new_size < _BUFFER_MAX_GROW:
            new_size = _BUFFER_MAX_GROW
        else:
            # Add a little extra
            new_size += _BUFFER_INITIAL_SIZE

        if self._smallbuf_inuse:
            new_buf = <char*> PyMem_Malloc(sizeof(char) * <size_t> new_size)
            if new_buf is NULL:  # pragma: nocover
                self._buf = NULL
                self._size = 0
                self._length = 0
                raise MemoryError
            memcpy(new_buf, self._buf, <size_t> self._size)
            self._size = new_size
            self._buf = new_buf
            self._smallbuf_inuse = False
        else:
            new_buf = <char*> PyMem_Realloc(<void*> self._buf, <size_t> new_size)
            if new_buf is NULL:
                PyMem_Free(self._buf)
                self._buf = NULL
                self._size = 0
                self._length = 0
                raise MemoryError
            self._buf = new_buf
            self._size = new_size

    cdef int write_buffer(self, WriteBuffer buf) except -1:
        if not buf._length:
            return 0

        self.ensure_allocated(buf._length)
        memcpy(self._buf + self._length,
               <void*> buf._buf,
               <size_t> buf._length)
        self._length += buf._length

    cdef int write_header(self, uint64_t sync,
                          tarantool.iproto_type op,
                          int64_t schema_id,
                          uint64_t stream_id) except -1:
        cdef:
            char *begin = NULL
            char *p = NULL
            uint32_t map_size
        self.ensure_allocated(HEADER_CONST_LEN)

        map_size = 2 \
                    + (<uint32_t> (schema_id > 0)) \
                    + (<uint32_t> (stream_id > 0))

        p = begin = &self._buf[self._length]
        p = mp_encode_map(&p[5], map_size)
        p = mp_encode_uint(p, tarantool.IPROTO_REQUEST_TYPE)
        p = mp_encode_uint(p, <uint32_t> op)
        p = mp_encode_uint(p, tarantool.IPROTO_SYNC)
        p = mp_encode_uint(p, sync)

        if schema_id > 0:  # pragma: nocover  # asynctnt does not send schema_id
            p = mp_encode_uint(p, tarantool.IPROTO_SCHEMA_VERSION)
            p = mp_store_u8(p, 0xce)
            p = mp_store_u32(p, schema_id)

        if stream_id > 0:
            p = mp_encode_uint(p, tarantool.IPROTO_STREAM_ID)
            p = mp_encode_uint(p, stream_id)

        self._length += (p - begin)

    cdef void write_length(self):
        cdef:
            char *p
        p = self._buf
        p = mp_store_u8(p, 0xce)
        p = mp_store_u32(p, self._length - 5)

    cdef char *mp_encode_nil(self, char *p) except NULL:
        cdef char *begin
        p = begin = self._ensure_allocated(p, 1)
        p = mp_encode_nil(p)
        self._length += (p - begin)
        return p

    cdef char *mp_encode_bool(self, char *p, bint value) except NULL:
        cdef char *begin
        p = begin = self._ensure_allocated(p, 1)
        p = mp_encode_bool(p, value)
        self._length += (p - begin)
        return p

    cdef char *mp_encode_double(self, char *p, double value) except NULL:
        cdef char *begin
        p = begin = self._ensure_allocated(p, mp_sizeof_double(value))
        p = mp_encode_double(p, value)
        self._length += (p - begin)
        return p

    cdef char *mp_encode_uint(self, char *p, uint64_t value) except NULL:
        cdef char *begin
        p = begin = self._ensure_allocated(p, mp_sizeof_uint(value))
        p = mp_encode_uint(p, value)
        self._length += (p - begin)
        return p

    cdef char *mp_encode_int(self, char *p, int64_t value) except NULL:
        cdef char *begin
        p = begin = self._ensure_allocated(p, mp_sizeof_int(value))
        p = mp_encode_int(p, value)
        self._length += (p - begin)
        return p

    cdef char *mp_encode_str(self, char *p,
                             const char *str, uint32_t len) except NULL:
        cdef char *begin
        p = begin = self._ensure_allocated(p, mp_sizeof_str(len))
        p = mp_encode_str(p, str, len)
        self._length += (p - begin)
        return p

    cdef char *mp_encode_bin(self, char *p,
                             const char *data, uint32_t len) except NULL:
        cdef char *begin
        p = begin = self._ensure_allocated(p, mp_sizeof_bin(len))
        p = mp_encode_bin(p, data, len)
        self._length += (p - begin)
        return p

    cdef char *mp_encode_decimal(self, char *p, object value) except NULL:
        cdef:
            char *begin
            uint8_t sign
            tuple digits
            uint32_t digits_count
            int exponent
            uint32_t length

        decimal_tuple = value.as_tuple()
        sign = <uint8_t> decimal_tuple.sign
        digits = <tuple> decimal_tuple.digits
        exponent = <int> decimal_tuple.exponent

        digits_count = <uint32_t> len(digits)
        length = decimal_len(exponent, digits_count)

        p = begin = self._ensure_allocated(p, mp_sizeof_ext(length))

        # encode header
        p = mp_encode_extl(p, tarantool.MP_DECIMAL, length)

        # encode decimal
        p = decimal_encode(p, digits_count, sign, digits, exponent)

        self._length += (p - begin)
        return p

    cdef char *mp_encode_uuid(self, char *p, object value) except NULL:
        cdef:
            char *begin
            char *data_p

        p = begin = self._ensure_allocated(p, mp_sizeof_ext(16))
        data_p = cpython.bytes.PyBytes_AS_STRING(<bytes> value.bytes)

        # encode header
        p = mp_encode_ext(p, tarantool.MP_UUID, data_p, 16)  # uuid is exactly 16 bytes

        self._length += (p - begin)
        return p

    cdef char *mp_encode_datetime(self, char *p, object value) except NULL:
        cdef:
            char *begin
            uint32_t length
            datetime pydt
            IProtoDateTime dt

        pydt = <datetime> value
        datetime_zero(&dt)
        datetime_from_py(pydt, &dt)

        length = datetime_len(&dt)
        p = begin = self._ensure_allocated(p, mp_sizeof_ext(length))
        p = mp_encode_extl(p, tarantool.MP_DATETIME, length)
        p = datetime_encode(p, &dt)
        self._length += (p - begin)
        return p

    cdef char *mp_encode_array(self, char *p, uint32_t len) except NULL:
        cdef char *begin
        p = begin = self._ensure_allocated(p, mp_sizeof_array(len))
        p = mp_encode_array(p, len)
        self._length += (p - begin)
        return p

    cdef char *mp_encode_map(self, char *p, uint32_t len) except NULL:
        cdef char *begin
        p = begin = self._ensure_allocated(p, mp_sizeof_map(len))
        p = mp_encode_map(p, len)
        self._length += (p - begin)
        return p

    cdef char *mp_encode_list(self, char *p, list arr) except NULL:
        cdef:
            uint32_t arr_len
            PyObject *item_ptr
            object item

        if arr is not None:
            arr_len = <uint32_t> cpython.list.PyList_GET_SIZE(arr)
        else:  # pragma: nocover
            arr_len = 0
        p = self.mp_encode_array(p, arr_len)

        if arr_len > 0:
            for i in range(arr_len):
                item_ptr = cpython.list.PyList_GET_ITEM(arr, i)
                item = <object> item_ptr
                p = self.mp_encode_obj(p, item)
        return p

    cdef char *mp_encode_tuple(self, char *p, tuple t) except NULL:
        cdef:
            uint32_t t_len
            PyObject *item_ptr
            object item

        if t is not None:
            t_len = <uint32_t> cpython.tuple.PyTuple_GET_SIZE(t)
        else:  # pragma: nocover
            t_len = 0
        p = self.mp_encode_array(p, t_len)

        if t_len > 0:
            for i in range(t_len):
                item_ptr = cpython.tuple.PyTuple_GET_ITEM(t, i)
                item = <object> item_ptr
                p = self.mp_encode_obj(p, item)
        return p

    cdef char *mp_encode_dict(self, char *p, dict d) except NULL:
        cdef:
            uint32_t d_len
            PyObject *pkey
            PyObject *pvalue
            object key, value
            Py_ssize_t pos

        if d is not None:
            d_len = <uint32_t> cpython.dict.PyDict_Size(d)
        else:  # pragma: nocover
            d_len = 0
        p = self.mp_encode_map(p, d_len)

        pos = 0
        while cpython.dict.PyDict_Next(d, &pos, &pkey, &pvalue):
            key = <object> pkey
            value = <object> pvalue
            p = self.mp_encode_obj(p, key)
            p = self.mp_encode_obj(p, value)

        return p

    cdef char *mp_encode_obj(self, char *p, object o) except NULL:
        cdef:
            bytes o_string_temp
            char *o_string_str
            ssize_t o_string_len

        if o is None:
            return self.mp_encode_nil(p)

        elif isinstance(o, float):
            return self.mp_encode_double(p, <double> o)

        elif isinstance(o, bool):
            return self.mp_encode_bool(p, <bint> o)

        elif isinstance(o, int):
            if o >= 0:
                return self.mp_encode_uint(p, <uint64_t> o)
            else:
                return self.mp_encode_int(p, <int64_t> o)

        elif isinstance(o, bytes):
            o_string_str = NULL
            o_string_len = 0
            cpython.bytes.PyBytes_AsStringAndSize(o,
                                                  &o_string_str,
                                                  &o_string_len)
            return self.mp_encode_bin(p, o_string_str, <uint32_t> o_string_len)

        elif isinstance(o, str):
            o_string_temp = encode_unicode_string(o, self._encoding)
            o_string_str = NULL
            o_string_len = 0
            cpython.bytes.PyBytes_AsStringAndSize(o_string_temp,
                                                  &o_string_str,
                                                  &o_string_len)

            p = self.mp_encode_str(p, o_string_str, <uint32_t> o_string_len)
            return p

        elif isinstance(o, list):
            return self.mp_encode_list(p, <list> o)

        elif isinstance(o, tuple):
            return self.mp_encode_tuple(p, <tuple> o)

        elif isinstance(o, dict):
            return self.mp_encode_dict(p, <dict> o)

        elif isinstance(o, datetime):
            return self.mp_encode_datetime(p, o)

        elif isinstance(o, Decimal):
            return self.mp_encode_decimal(p, o)

        elif isinstance(o, UUID):
            return self.mp_encode_uuid(p, o)

        else:
            raise TypeError(
                'Type `{}` is not supported for encoding'.format(type(o)))



cimport cython
from libc.stdint cimport int32_t, int64_t, uint32_t, uint64_t


cdef struct Header:
    int32_t code
    int32_t return_code
    uint64_t sync
    int64_t schema_id

cdef class IProtoErrorStackFrame:
    cdef:
        readonly str error_type
        readonly str file
        readonly int line
        readonly str message
        readonly int err_no
        readonly int code
        readonly dict fields

cdef class IProtoError:
    cdef:
        readonly list trace

cdef class Response:
    cdef:
        int32_t code_
        int32_t return_code_
        uint64_t sync_
        int64_t schema_id_
        readonly str errmsg
        readonly IProtoError error
        int _rowcount
        readonly list body
        readonly bytes encoding
        readonly Metadata metadata
        readonly Metadata params
        readonly int params_count
        readonly list autoincrement_ids
        uint64_t stmt_id_
        bint _push_subscribe
        BaseRequest request_
        object _exception

        readonly object _q
        readonly object _push_event
        object _q_append
        object _q_popleft
        object _push_event_set
        object _push_event_clear

    cdef inline bint is_error(self)
    cdef inline uint32_t _len(self)
    cdef inline void init_push(self)
    cdef inline void add_push(self, push)
    cdef inline object pop_push(self)
    cdef inline int push_len(self)
    cdef inline void set_data(self, list data)
    cdef inline void set_exception(self, exc)
    cdef inline object get_exception(self)
    cdef inline void notify(self)

cdef ssize_t response_parse_header(const char *buf, uint32_t buf_len,
                                   Header *hdr) except -1
cdef ssize_t response_parse_body(const char *buf, uint32_t buf_len,
                                 Response resp, BaseRequest req,
                                 bint is_chunk) except -1
cdef IProtoError parse_iproto_error(const char ** b, bytes encoding)



from libc.stdint cimport int64_t, uint64_t


cdef class BaseRequest:
    cdef:
        tarantool.iproto_type op
        uint64_t sync
        int64_t schema_id
        uint64_t stream_id
        SchemaSpace space
        object waiter
        object timeout_handle
        bint parse_metadata
        bint parse_as_tuples
        bint push_subscribe
        bint check_schema_change

    cdef inline Metadata metadata(self):
        if self.space is None:
            return None
        return self.space.metadata

    cdef inline WriteBuffer encode(self, bytes encoding)
    cdef int encode_body(self, WriteBuffer buffer) except -1


cdef char *encode_key_sequence(WriteBuffer buffer,
                               char *p, object t,
                               Metadata metadata,
                               bint default_none) except NULL



cimport cpython.unicode
from cpython.ref cimport PyObject


cdef bytes encode_unicode_string(object s, bytes encoding):
    cdef:
        bytes b
        PyObject *p

    b = <bytes><object>cpython.unicode.PyUnicode_AsEncodedString(
        s, encoding, b'strict'
    )
    return b


cdef str decode_string(bytes b, bytes encoding):
    return <str><object>cpython.unicode.PyUnicode_FromEncodedObject(
        b, encoding, b'strict'
    )



from asynctnt import Response
from asynctnt.exceptions import TarantoolSchemaError
from tests import BaseTarantoolTestCase


class UpsertTestCase(BaseTarantoolTestCase):
    async def _fill_data(self):
        data = [
            [0, "a", 1],
            [1, "b", 0],
        ]
        for t in data:
            await self.conn.insert(self.TESTER_SPACE_ID, t)

        return data

    async def test__upsert_empty_one_assign(self):
        data = [0, "hello2", 1, 4, "what is up"]

        res = await self.conn.upsert(self.TESTER_SPACE_ID, data, [["=", 2, 2]])
        self.assertIsInstance(res, Response, "Got response")
        self.assertEqual(res.code, 0, "success")
        self.assertGreater(res.sync, 0, "sync > 0")
        self.assertResponseEqual(res, [], "Body ok")

        res = await self.conn.select(self.TESTER_SPACE_ID, [0])
        self.assertResponseEqual(res, [data], "Body ok")

    async def test__upsert_update_one_assign(self):
        data = [0, "hello2", 1, 4, "what is up"]

        await self.conn.insert(self.TESTER_SPACE_ID, data)
        res = await self.conn.upsert(self.TESTER_SPACE_ID, data, [["=", 2, 2]])
        self.assertIsInstance(res, Response, "Got response")
        self.assertEqual(res.code, 0, "success")
        self.assertGreater(res.sync, 0, "sync > 0")
        self.assertResponseEqual(res, [], "Body ok")

        res = await self.conn.select(self.TESTER_SPACE_ID, [0])
        data[2] = 2
        self.assertResponseEqual(res, [data], "Body ok")

    async def test__upsert_by_name(self):
        data = [0, "hello2", 1, 4, "what is up"]

        await self.conn.upsert(self.TESTER_SPACE_NAME, data, [["=", 2, 2]])

        res = await self.conn.select(self.TESTER_SPACE_ID, [0])
        self.assertIsInstance(res, Response, "Got response")
        self.assertEqual(res.code, 0, "success")
        self.assertGreater(res.sync, 0, "sync > 0")
        self.assertResponseEqual(res, [data], "Body ok")

    async def test__upsert_by_name_no_schema(self):
        await self.tnt_reconnect(fetch_schema=False)

        with self.assertRaises(TarantoolSchemaError):
            await self.conn.upsert(
                self.TESTER_SPACE_NAME, [0, "hello", 1], [["=", 2, 2]]
            )

    async def test__upsert_dict_key(self):
        data = {
            "f1": 0,
            "f2": "hello",
            "f3": 1,
            "f4": 2,
            "f5": 100,
        }

        res = await self.conn.upsert(self.TESTER_SPACE_ID, data, [["=", 2, 2]])
        self.assertResponseEqual(res, [], "Body ok")

        res = await self.conn.select(self.TESTER_SPACE_ID, [0])
        self.assertResponseEqual(res, [[0, "hello", 1, 2, 100]], "Body ok")

    async def test__usert_dict_resp_no_effect(self):
        data = {
            "f1": 0,
            "f2": "hello",
            "f3": 1,
            "f4": 10,
            "f5": 1000,
        }

        res = await self.conn.upsert(self.TESTER_SPACE_ID, data, [["=", 2, 2]])
        self.assertResponseEqual(res, [], "Body ok")



cimport cpython
cimport cpython.bytes
cimport cython

import hashlib


cdef inline bytes _sha1(tuple values):
    cdef object sha = hashlib.sha1()
    for i in values:
        if i is not None:
            sha.update(i)
    return sha.digest()

cdef inline bytes _strxor(bytes hash1, bytes scramble):
    cdef:
        char *hash1_str
        ssize_t hash1_len

        char *scramble_str
        ssize_t scramble_len
    cpython.bytes.PyBytes_AsStringAndSize(hash1,
                                          &hash1_str, &hash1_len)
    cpython.bytes.PyBytes_AsStringAndSize(scramble,
                                          &scramble_str, &scramble_len)
    for i in range(scramble_len):
        scramble_str[i] = hash1_str[i] ^ scramble_str[i]
    return scramble

@cython.final
cdef class AuthRequest(BaseRequest):
    cdef int encode_body(self, WriteBuffer buffer) except -1:
        cdef:
            char *begin
            char *p
            uint32_t body_map_sz
            uint32_t max_body_len

            char *username_str
            ssize_t username_len

            char *scramble_str
            ssize_t scramble_len

        username_bytes = encode_unicode_string(self.username, buffer._encoding)
        password_bytes = encode_unicode_string(self.password, buffer._encoding)

        hash1 = _sha1((password_bytes,))
        hash2 = _sha1((hash1,))
        scramble = _sha1((self.salt, hash2))
        scramble = _strxor(hash1, scramble)

        cpython.bytes.PyBytes_AsStringAndSize(username_bytes,
                                              &username_str, &username_len)
        cpython.bytes.PyBytes_AsStringAndSize(scramble,
                                              &scramble_str, &scramble_len)
        body_map_sz = 2
        # Size description:
        # mp_sizeof_map()
        # + mp_sizeof_uint(TP_USERNAME)
        # + mp_sizeof_str(username_len)
        # + mp_sizeof_uint(TP_TUPLE)
        # + mp_sizeof_array(2)
        # + mp_sizeof_str(9) (chap-sha1)
        # + mp_sizeof_str(SCRAMBLE_SIZE)
        max_body_len = 1 \
                       + 1 \
                       + mp_sizeof_str(<uint32_t> username_len) \
                       + 1 \
                       + 1 \
                       + 1 + 9 \
                       + mp_sizeof_str(<uint32_t> scramble_len)

        buffer.ensure_allocated(max_body_len)

        p = begin = &buffer._buf[buffer._length]
        p = mp_encode_map(p, body_map_sz)
        p = mp_encode_uint(p, tarantool.IPROTO_USER_NAME)
        p = mp_encode_str(p, username_str, <uint32_t> username_len)

        p = mp_encode_uint(p, tarantool.IPROTO_TUPLE)
        p = mp_encode_array(p, 2)
        p = mp_encode_str(p, "chap-sha1", 9)
        p = mp_encode_str(p, scramble_str, <uint32_t> scramble_len)
        buffer._length += (p - begin)



# cython: language_level=3

cimport cpython.dict
from cpython.datetime cimport import_datetime

import_datetime()

import asyncio
import enum

from asynctnt.exceptions import TarantoolNotConnectedError

include "const.pxi"

include "unicodeutil.pyx"
include "schema.pyx"
include "ext.pyx"
include "buffer.pyx"
include "rbuffer.pyx"

include "requests/base.pyx"
include "requests/ping.pyx"
include "requests/call.pyx"
include "requests/eval.pyx"
include "requests/select.pyx"
include "requests/insert.pyx"
include "requests/delete.pyx"
include "requests/update.pyx"
include "requests/upsert.pyx"
include "requests/prepare.pyx"
include "requests/execute.pyx"
include "requests/id.pyx"
include "requests/auth.pyx"
include "requests/streams.pyx"

include "ttuple.pyx"
include "response.pyx"
include "db.pyx"
include "push.pyx"

include "coreproto.pyx"


class Iterator(enum.IntEnum):
    """
        Available Iterator types
    """
    EQ = 0
    REQ = 1
    ALL = 2
    LT = 3
    LE = 4
    GE = 5
    GT = 6
    BITS_ALL_SET = 7
    BITS_ANY_SET = 8
    BITS_ALL_NOT_SET = 9
    OVERLAPS = 10
    NEIGHBOR = 11


cdef class BaseProtocol(CoreProtocol):
    def __init__(self, host, port,
                 username, password,
                 fetch_schema,
                 auto_refetch_schema,
                 connected_fut,
                 on_connection_made, on_connection_lost,
                 loop,
                 request_timeout=None,
                 encoding=None,
                 initial_read_buffer_size=None):
        CoreProtocol.__init__(self, host, port, encoding,
                              initial_read_buffer_size)

        self.loop = loop

        self.username = username
        self.password = password
        self.fetch_schema = fetch_schema
        self.auto_refetch_schema = auto_refetch_schema
        self.request_timeout = request_timeout or 0
        self.post_con_state = POST_CONNECTION_NONE

        self.connected_fut = connected_fut
        self.on_connection_made_cb = on_connection_made
        self.on_connection_lost_cb = on_connection_lost
        self._closing = False

        self._on_request_completed_cb = self._on_request_completed
        self._on_request_timeout_cb = self._on_request_timeout

        self._reqs = {}
        self._sync = 0
        self._last_stream_id = 0
        self._schema_id = -1
        self._schema = Schema.__new__(Schema, self._schema_id)
        self._schema_fetch_in_progress = False
        self._refetch_schema_future = None
        self._db = self._create_db(<bint> False)
        self.execute = self._execute_bad

        try:
            self.create_future = self.loop.create_future
        except AttributeError:
            self.create_future = self._create_future_fallback

    def _create_future_fallback(self):  # pragma: nocover
        return asyncio.Future(loop=self.loop)

    @property
    def schema(self):
        return self._schema

    @property
    def schema_id(self):
        return self._schema_id

    cdef void _set_connection_ready(self):
        if not self.connected_fut.done():
            self.connected_fut.set_result(True)
            self.con_state = CONNECTION_FULL

    cdef void _set_connection_error(self, e):
        if not self.connected_fut.done():
            self.connected_fut.set_exception(e)
            self.con_state = CONNECTION_BAD
            self.post_con_state = POST_CONNECTION_NONE
            self.execute = self._execute_bad

    cdef void _on_greeting_received(self):
        self.post_con_state = POST_CONNECTION_ID
        self.execute = self._execute_normal
        self._post_con_state_machine()

    cdef void _post_con_state_machine(self):
        if self.post_con_state == POST_CONNECTION_ID:
            assert self.version is not None
            if self.version >= (2, 10, 0):
                # send <id> request
                self._do_id()
                return
            else:
                # Tarantool does not support id call - call auth request
                self.post_con_state = POST_CONNECTION_AUTH

        if self.post_con_state == POST_CONNECTION_AUTH:
            if self.username and self.password:
                # send <auth> request
                self._do_auth(self.username, self.password)
                return
            else:
                # no need to auth - fetch schema
                self.post_con_state = POST_CONNECTION_SCHEMA

        if self.post_con_state == POST_CONNECTION_SCHEMA:
            if self.fetch_schema:
                self._refetch_schema()
                return
            else:
                # nothing else to do - connection is done
                self.post_con_state = POST_CONNECTION_DONE

        if self.post_con_state == POST_CONNECTION_DONE:
            self._set_connection_ready()
            return

    cdef void _on_response_received(self, const char *buf, uint32_t buf_len):
        cdef:
            PyObject *req_p
            Response response
            BaseRequest req
            Header hdr
            bint is_chunk
            object waiter
            object sync_obj
            object err

            ssize_t length

        length = response_parse_header(buf, buf_len, &hdr)
        buf_len -= length
        buf = &buf[length]  # skip header

        sync_obj = <object> hdr.sync

        response_p = cpython.dict.PyDict_GetItem(self._reqs, sync_obj)
        if response_p is NULL:
            logger.warning('sync %d not found', hdr.sync)
            return

        is_chunk = (hdr.code == tarantool.IPROTO_CHUNK)

        response = <Response> response_p
        req = response.request_
        response.sync_ = hdr.sync
        response.schema_id_ = hdr.schema_id
        if not is_chunk:
            response.code_ = hdr.code
            response.return_code_ = hdr.return_code
            cpython.dict.PyDict_DelItem(self._reqs, sync_obj)
        else:
            if not req.push_subscribe:
                # skip request data as no one will be waiting for it
                mp_next(&buf)
                return

        err = None
        if buf != &buf[buf_len]:
            # has body
            try:
                response_parse_body(buf, buf_len, response, req, is_chunk)
            except Exception as e:
                err = e

        # refetch schema if it is changed
        if self.con_state == CONNECTION_FULL \
                and req.check_schema_change \
                and self.auto_refetch_schema \
                and response.schema_id_ > 0 \
                and response.schema_id_ != self._schema_id:
            self._refetch_schema()

        # returning result
        if is_chunk:
            return

        waiter = req.waiter
        if waiter is None or waiter.done():
            return

        if err is not None:
            response.set_exception(err)
            waiter.set_exception(err)
            return

        if response.is_error():
            err = TarantoolDatabaseError(response.return_code_,
                                         response.errmsg,
                                         response.error)
            response.set_exception(err)
            waiter.set_exception(err)
            return

        waiter.set_result(response)

    cdef void _do_id(self):
        fut = self._db._id(0.0)

        def on_id(f):
            if f.cancelled():
                self._set_connection_error(asyncio.futures.CancelledError())
                return
            e = f.exception()
            if not e:
                logger.debug('Tarantool[%s:%s] identified successfully',
                             self.host, self.port)

                self.post_con_state = POST_CONNECTION_AUTH
                self._post_con_state_machine()
            else:
                logger.error('Tarantool[%s:%s] identification failed: %s',
                             self.host, self.port, str(e))
                self._set_connection_error(e)

        fut.add_done_callback(on_id)

    cdef void _do_auth(self, str username, str password):
        # No extra error handling from Db.execute
        fut = self._db._auth(self.salt, username, password, 0)

        def on_authorized(f):
            if f.cancelled():
                self._set_connection_error(asyncio.futures.CancelledError())
                return
            e = f.exception()
            if not e:
                logger.debug('Tarantool[%s:%s] Authorized successfully',
                             self.host, self.port)

                self.post_con_state = POST_CONNECTION_SCHEMA
                self._post_con_state_machine()
            else:
                logger.error('Tarantool[%s:%s] Authorization failed: %s',
                             self.host, self.port, str(e))
                self._set_connection_error(e)

        fut.add_done_callback(on_authorized)

    cdef void _do_fetch_schema(self, object fut):
        if self._schema_fetch_in_progress:
            return

        self._schema_fetch_in_progress = True

        def on_fetch(f):
            self._schema_fetch_in_progress = False

            if f.cancelled():
                self._set_connection_error(asyncio.futures.CancelledError())
                return
            e = f.exception()
            if not e:
                if fut is not None and fut.cancelled():
                    # if the caller has cancelled waiting
                    return

                spaces, indexes = f.result()
                logger.debug('Tarantool[%s:%s] Schema fetch succeeded. '
                             'Version: %d, Spaces: %d, Indexes: %d.',
                             self.host, self.port,
                             spaces.schema_id, len(spaces), len(indexes))
                try:
                    self._schema = Schema.parse(spaces.schema_id,
                                                spaces, indexes)
                except Exception as e:
                    logger.exception(e)
                    logger.error('Error happened while parsing schema. '
                                 'Space, fields and index names currently '
                                 'not working. Please file an issue at '
                                 'https://github.com/igorcoding/asynctnt')
                    self.auto_refetch_schema = False
                    self.fetch_schema = False
                    self._schema_id = -1
                    self._set_connection_ready()
                    if fut is not None and not fut.done():
                        fut.set_result(None)
                    return

                self._schema_id = self._schema.id
                self.post_con_state = POST_CONNECTION_DONE
                self._post_con_state_machine()
                if fut is not None and not fut.done():
                    fut.set_result(self._schema)
            else:
                if self._closing:
                    # show a diag message rather than Lost connection to Tarantool when disconnected (#19)
                    logger.debug('Schema fetch stopped: connection is closed')
                    return

                logger.error('Tarantool[%s:%s] Schema fetch failed: %s',
                             self.host, self.port, str(e))
                if isinstance(e, asyncio.TimeoutError):
                    e = asyncio.TimeoutError('Schema fetch timeout')
                self._set_connection_error(e)
                if fut is not None and not fut.done():
                    fut.set_exception(e)

        fut_vspace = self._db.select(SPACE_VSPACE, timeout=0,
                                     check_schema_change=False)
        fut_vindex = self._db.select(SPACE_VINDEX, timeout=0,
                                     check_schema_change=False)
        gather_fut = asyncio.gather(fut_vspace, fut_vindex,
                                    return_exceptions=False)
        gather_fut.add_done_callback(on_fetch)

    cdef void _on_connection_made(self):
        CoreProtocol._on_connection_made(self)

        if self.on_connection_made_cb:
            self.on_connection_made_cb()

    cdef void _on_connection_lost(self, exc):
        cdef:
            BaseRequest req
            Response response
            PyObject *pkey
            PyObject *pvalue
            object key, value
            Py_ssize_t pos

        if self._closing:
            return

        self._closing = True
        self.post_con_state = POST_CONNECTION_NONE
        self.execute = self._execute_bad

        pos = 0
        while cpython.dict.PyDict_Next(self._reqs, &pos, &pkey, &pvalue):
            sync = <uint64_t> <object> pkey
            response = <Response> pvalue
            req = response.request_

            waiter = req.waiter
            if waiter and not waiter.done():
                err = None
                if exc is None:
                    err = TarantoolNotConnectedError(
                        'Lost connection to Tarantool'
                    )
                elif isinstance(exc, (ConnectionRefusedError,
                                      ConnectionResetError)):
                    err = TarantoolNotConnectedError(
                        'Lost connection to Tarantool: {}: {}'.format(
                            exc.__class__.__name__, str(exc))
                    )
                elif exc is ConnectionRefusedError \
                        or exc is ConnectionResetError:
                    err = TarantoolNotConnectedError(
                        'Lost connection to Tarantool: {}'.format(
                            exc.__name__)
                    )
                else:
                    err = exc

                if err is not None:
                    waiter.set_exception(err)
                    if response is not None:
                        response.set_exception(err)

        if self.on_connection_lost_cb:
            self.on_connection_lost_cb(exc)

        self._reqs = {}  # reset requests map

    cdef inline uint64_t next_sync(self):
        self._sync += 1
        return self._sync

    cdef inline uint64_t next_stream_id(self):
        self._last_stream_id += 1
        return self._last_stream_id

    def _on_request_timeout(self, waiter):
        cdef:
            BaseRequest req
            Response response

        if waiter.done():
            return

        response = waiter._response
        req = response.request_
        req.timeout_handle.cancel()
        req.timeout_handle = None
        waiter.set_exception(
            asyncio.TimeoutError(
                '{} exceeded timeout'.format(req.__class__.__name__))
        )

    def _on_request_completed(self, fut):
        cdef BaseRequest req = (<Response> fut._response).request_
        fut._response = None

        if req.timeout_handle is not None:
            req.timeout_handle.cancel()
            req.timeout_handle = None

    cdef object _new_waiter_for_request(self, Response response, BaseRequest req, float timeout):
        fut = self.create_future()
        req.waiter = fut
        fut._response = response  # to be able to retrieve request after done()

        if timeout < 0:
            timeout = self.request_timeout
        if timeout is not None and timeout > 0:
            req.timeout_handle = \
                self.loop.call_later(timeout,
                                     self._on_request_timeout_cb, fut)
            fut.add_done_callback(self._on_request_completed_cb)
        return fut

    cdef Db _create_db(self, bint gen_stream_id):
        cdef uint64_t stream_id
        if gen_stream_id:
            stream_id = self.next_stream_id()
        else:
            stream_id = 0

        return Db.create(self, stream_id)

    def create_db(self, bint gen_stream_id = False):
        return self._create_db(gen_stream_id)

    def get_common_db(self):
        return self._db

    cdef object _execute_bad(self, BaseRequest req, float timeout):
        raise TarantoolNotConnectedError('Tarantool is not connected')

    cdef object _execute_normal(self, BaseRequest req, float timeout):
        cdef Response response
        response = <Response> Response.__new__(Response)
        response.request_ = req
        response.encoding = self.encoding
        if req.push_subscribe:
            response.init_push()
        cpython.dict.PyDict_SetItem(self._reqs, req.sync, response)
        self._write(req.encode(self.encoding))

        return self._new_waiter_for_request(response, req, timeout)

    cdef uint32_t transform_iterator(self, iterator) except *:
        if isinstance(iterator, int):
            return iterator
        if isinstance(iterator, Iterator):
            return iterator.value
        if isinstance(iterator, str):
            return Iterator[iterator]
        else:
            raise TypeError('Iterator is of unsupported type '
                            '(asynctnt.Iterator, int, str)')

    cdef object _refetch_schema(self):
        if self._refetch_schema_future is not None and not self._refetch_schema_future.done():
            self._schema_fetch_in_progress = False
            self._refetch_schema_future.cancel()

        self._refetch_schema_future = self.create_future()
        self._do_fetch_schema(self._refetch_schema_future)

        return self._refetch_schema_future

    def refetch_schema(self):
        return self._refetch_schema()


class Protocol(BaseProtocol, asyncio.Protocol):
    pass

